import os, sys, argparse, re
from xutil import log


def pykill(*args_):
  "Kill processes based on name"

  import psutil

  args = sys.argv[1:] or args_
  if not args:
    log('~Need one argument to kill processes!')
    return

  kw = args[0]
  user = os.getenv('USER')
  skip_list = set(['login'])
  this_pid = os.getpid()

  procs_matched = []
  for proc in psutil.process_iter():
    pid = proc.pid
    name = proc.name()

    if not kw.strip(
    ) or user != proc.username() or name in skip_list or pid == this_pid:
      continue
    cmdline = ' '.join(proc.cmdline())

    try:
      if kw.lower() in cmdline.lower():
        proc.terminate()
        log('+Killed {} : {}'.format(pid, cmdline))
        procs_matched.append(proc)
    except Exception as E:
      log('-Could not kill {} : {}'.format(pid, name))
      log(E)

    if len(procs_matched) == 0:
      log('-Did not match any process name.')


def create_profile():
  "Create profile.yaml if it does not exists"
  from xutil.helpers import get_profile
  get_profile(create_if_missing=True)
  log('+YAML Profile located @ {}'.format(os.environ['PROFILE_YAML']))



def alias_cli():
  "Install alias"
  from xutil.helpers import get_home_path, get_dir_path, get_script_path
  from xutil.diskio import read_file, write_file
  from shutil import copyfile
  ans = input("Install 'alias.sh' in home directory (Y to proceed)? ")
  if ans.lower() != 'y':
    return

  src_path = get_dir_path() + '/alias.sh'
  dst_path = get_home_path() + '/.xutil.alias.sh'
  bash_profile_path = get_home_path() + '/.bashrc'

  # log('src_path -> ' + src_path)
  # log('dst_path -> ' + dst_path)
  copyfile(src_path, dst_path)

  bash_prof_text = read_file(bash_profile_path)

  if not dst_path in bash_prof_text:
    bash_prof_text = '{}\n\n. {}\n'.format(bash_prof_text, dst_path)
    write_file(bash_profile_path, bash_prof_text)
    log('+Updated ' + bash_profile_path)


def exec_sql():
  from xutil.database.etl import SqlCmdParser
  SqlCmdParser()

def profile_entry():
  from xutil.helpers import load_profile, save_profile
  from collections import OrderedDict
  import urllib.parse as urlparse
  import yaml


  parser = argparse.ArgumentParser(description='Profile Entry Maker')
  parser.add_argument('--url', help='the URL')
  parser.add_argument('--name', help='Database Name')
  parser.add_argument('--url_type', help='Type of URL (SQL Alchemy is default)')
  parser.add_argument('--add', action='store_true', help='Whether to add to current profile.yaml')
  args = parser.parse_args()

  if not (args.url and args.name):
    raise Exception("Need to provide --url and --name")
  
  url_vars = urlparse.urlparse(args.url)

  prof = {}

  prof['name'] = args.name
  prof['host'] = url_vars.hostname
  prof['port'] = url_vars.port
  prof['user'] = url_vars.username
  prof['password'] = url_vars.password
  prof['database'] = url_vars.path[1:]
  prof['type'] = url_vars.scheme

  if 'postgres' in prof['type']:
    prof['type'] = 'postgresql'
  if 'mssql' in prof['type']:
    prof['type'] = 'sqlserver'
  if 'oracle' in prof['type']:
    prof['type'] = 'oracle'
  
  if prof['type'] in ('postgresql', 'redshift'):
    prof['sslmode'] = 'require'
    prof['url'] = f"jdbc:postgresql://{prof['host']}:{prof['port']}/{prof['database']}"
    prof['sa_url'] = args.url

  if args.add:
    profile_yaml = load_profile()
    profile_yaml['databases'][args.name] = prof
    save_profile(profile_yaml)
    log('+ Set "{}" in profile.yaml'.format(args.name))
  else:
    print(yaml.dump({args.name : prof}))

def exec_etl():
  from xutil.database.etl import EtlCmdParser
  EtlCmdParser()


def ipy_imports(launch_spark=False):
  from xutil.helpers import (log, get_exception_message, get_error_str, get_profile)
  from xutil.database.base import get_conn
  from xutil.diskio import (write_csv, read_csv, read_file, write_file,
                            get_hdfs)
  from jmespath import search
  from pathlib import Path
  from collections import Counter, namedtuple, OrderedDict
  import time, datetime

  if launch_spark:
    from xutil.database.spark import Spark
    parser = argparse.ArgumentParser(description='Spark IPython')
    parser.add_argument(
      '--master', help='Master string for Spark Instance')
    parser.add_argument(
      '--profile', help='Database profile name from PROFILE_YAML')
    args = parser.parse_args()
    dbs = get_profile(create_if_missing=True)['databases']
    if args.profile and args.profile in dbs and dbs[args.profile]['type'].lower() in ('hive', 'spark'):
      conn = get_conn(args.profile)
      globals()['sparko'] = conn.sparko
    elif args.profile:
      log(Exception('Profile {} not found or incompatible.'.format(args.profile)))
      sys.exit(1)
    else:
      globals()['sparko'] = Spark(master=args.master)
    globals()['sc'] = sparko.sc
    globals()['spark'] = sparko.spark

  ldict = locals()
  for name in ldict:
    var = ldict[name]
    if callable(var) or isinstance(var, __builtins__.__class__):
      # is a function or class or module
      globals()[name] = var


def ipy(launch_spark=False):
  import IPython
  ipy_imports(launch_spark=launch_spark)
  IPython.embed(using=False) # https://github.com/ipython/ipython/issues/11523


def ipy_spark():
  ipy(launch_spark=True)


import os
from xutil.database.base import DBConn
from xutil.helpers import get_dir_path, log
from collections import namedtuple

# TODO:
# Add Metadata functions https://www.google.com/search?q=jaydebeapi+DatabaseMetaData
# getColumns
# getColumnPrivileges
# getMetaData
# getTypeInfo
# getTables
# getTableTypes

# meta = conn.connection.jconn.getMetaData()
# https://stackoverflow.com/a/21140430
# https://docs.oracle.com/javase/7/docs/api/java/sql/DatabaseMetaData.html


class JdbcConn(DBConn):
  """Any JDBC Connection String"""

  def connect(self):
    "Connect / Re-Connect to Database"
    import jaydebeapi

    cred = self._cred
    profile = self.profile
    jar_path = get_jar_path(cred.type, profile)

    os.environ['CLASSPATH'] = '{}:{}'.format(
      jar_path, os.environ['CLASSPATH'] if 'CLASSPATH' in os.environ else '')

    self.connection = jaydebeapi.connect(
      profile['drivers'][cred.type]['class'],
      cred.url,
      [cred.user, cred.password],
      jar_path,
    )
    self.cursor = None
    self.meta = self.get_meta()

    # self.connection.autocommit = True
    self.name = cred.type
    self.username = cred.user if cred else ''

    cursor = self.get_cursor()

  def set_variables(self):
    if self.type == 'sqlserver':
      self.truncate_datef = lambda f: 'CONVERT(DATE, {})'.format(f)
    if self.type == 'oracle':
      self.truncate_datef = lambda f: 'trunc({})'.format(f)

  def get_meta(self):
    return self.connection.jconn.getMetaData()

  def get_catalogs(self, echo=True):
    "Get list of Catalogs."

    Rec = namedtuple('Catalogs', 'catalog')
    rs_rows = rs_to_rows(self.meta.getCatalogs())
    self._fields = Rec._fields

    rows = [Rec(r[0]) for r in rs_rows]
    return rows


  def get_schemas(self, echo=True):
    "Get list of schemas."

    Rec = namedtuple('Schemas', 'schema')
    rs_rows = rs_to_rows(self.meta.getSchemas())

    self._fields = Rec._fields

    rows = [Rec(r[0]) for r in rs_rows]
    return rows


  def get_tables(self, schema, catalog=None, echo=True):
    "Get metadata for tables."
    Rec = namedtuple('Table', 'schema table')
    self._fields = Rec._fields

    rs_rows = rs_to_rows(self.meta.getTables(catalog, schema, None, ['TABLE']))

    rows = [Rec(schema, r.TABLE_NAME) for r in rs_rows]
    return rows

  def get_views(self, schema, catalog=None, echo=True):
    "Get metadata for views."
    Rec = namedtuple('View', 'schema view')
    self._fields = Rec._fields

    rs_rows = rs_to_rows(self.meta.getTables(catalog, schema, None, ['VIEW']))

    rows = [Rec(schema, r.TABLE_NAME) for r in rs_rows]
    return rows

  def get_columns(self,
                  table_name,
                  catalog=None,
                  object_type=None,
                  echo=False,
                  include_schema_table=True,
                  native_type=True):
    "Get column metadata for table"
    # getColumns(String catalog, String schemaPattern, String tableNamePattern, String columnNamePattern)
    if include_schema_table:
      headers = 'schema table id column_name type nullable default autoincrement'
    else:
      headers = 'id column_name type nullable default autoincrement'

    Rec = namedtuple('Columns', headers)
    self._fields = Rec._fields
    all_rows = []

    table_names = table_name if isinstance(table_name, list) else [table_name]

    for table_name in table_names:
      schema, table = self._split_schema_table(table_name)

      def get_rec(r_dict, column_order):
        if include_schema_table:
          r_dict['schema'] = schema
          r_dict['table'] = table

        r_dict['column_name'] = r_dict['COLUMN_NAME']
        r_dict['type'] = r_dict['TYPE_NAME']
        r_dict['nullable'] = r_dict['IS_NULLABLE']
        r_dict['autoincrement'] = r_dict['IS_AUTOINCREMENT']
        r_dict['default'] = r_dict['COLUMN_DEF']
        r_dict['id'] = column_order

        for k in list(r_dict):
          if k not in headers.split():
            del r_dict[k]

        if '(' in r_dict['type']:
          r_dict['type'] = r_dict['type'].split('(')[0]

        return Rec(**r_dict)

      rs_rows = rs_to_rows(self.meta.getColumns(catalog, schema, table, None))

      all_rows += [get_rec(r._asdict(), i + 1) for i, r in enumerate(rs_rows)]
    self._fields = Rec._fields
    return all_rows

  def get_primary_keys(self, table_name, echo=False):
    "Get PK metadata for table"
    Rec = namedtuple('PKs', 'schema table pk_name column_name column_order')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    # def get_rec(col, pk_name, column_order):
    #   r_dict = {}
    #   r_dict['schema'] = schema
    #   r_dict['table'] = table
    #   r_dict['pk_name'] = pk_name
    #   r_dict['column_name'] = col
    #   r_dict['column_order'] = column_order
    #   return Rec(**r_dict)

    # sql_tmpl = self._template('metadata.primary_keys')
    # if sql_tmpl:
    #   rows = self.select(sql_tmpl.format(table=table, schema=schema))
    # else:
    #   self.get_engine(echo=echo)
    #   r_dict = self.engine_inspect.get_pk_constraint(table, schema=schema)
    #   rows = [
    #     get_rec(col, r_dict['name'], i + 1)
    #     for i, col in enumerate(r_dict['constrained_columns'])
    #   ]
    # return rows

    # getPrimaryKeys(String catalog, String schema, String table)
    rs_rows = rs_to_rows(self.meta.getPrimaryKeys(catalog, schema, table))

    log(Exception('getPrimaryKeys not implemented!'))
    print(rs_rows)


  def get_indexes(self, table_name, echo=False):
    "Get indexes metadata for table"

    Rec = namedtuple(
      'Indexes', 'schema table index_name column_name column_order unique')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    # def get_rec(r_dict):
    #   r_dict['schema'] = schema
    #   r_dict['table'] = table
    #   r_dict['index_name'] = r_dict['name']
    #   r_dict['unique'] = str(r_dict['unique'])
    #   del r_dict['name']
    #   for i, col in enumerate(r_dict['column_names']):
    #     r_dict['column_name'] = col
    #     r_dict['column_order'] = i + 1
    #     yield Rec(**r_dict)

    # sql_tmpl = self._template('metadata.indexes')
    # if sql_tmpl:
    #   rows = self.select(sql_tmpl.format(table=table, schema=schema))
    # else:
    #   self.get_engine(echo=echo)
    #   rows = self.engine_inspect.get_indexes(table, schema=schema)
    #   rows = [get_rec(r_dict) for r_dict in rows]

    # return rows

    # getIndexInfo(String catalog, String schema, String table, boolean unique, boolean approximate)
    rs_rows = rs_to_rows(
      self.meta.getIndexInfo(catalog, schema, table, None, None))

    log(Exception('getIndexInfo not implemented!'))
    print(rs_rows)

  def get_ddl(self, table_name, object_type=None, echo=True):
    "Get ddl for table"
    Rec = namedtuple('DDL', 'ddl')
    self._fields = Rec._fields
    schema, table = self._split_schema_table(table_name)

    sql_tmpl = self._template('metadata.ddl')
    if sql_tmpl:
      rows = self.select(
        sql_tmpl.format(
          schema=schema,
          table=table,
          obj_type=object_type,
        ))
    else:
      log(Exception('get_ddl not implemented!'))

    return ['get_ddl not implemented!']


def get_jar_path(db_type, profile):
  from xutil.diskio import get_zip_path
  from zipfile import ZipFile
  from pathlib import Path
  import shutil

  tmp_folder = profile['variables']['tmp_folder']

  jar_path = profile['drivers'][db_type]['path']

  if Path(jar_path).exists():
    return jar_path

  # try relative path
  jar_path = '{}/{}'.format(
    get_dir_path(__file__), profile['drivers'][db_type]['path'])

  if '.zip' in jar_path.lower():
    new_jar_path = '{}/{}'.format(tmp_folder, Path(jar_path).name)
    if not Path(new_jar_path).exists():
      zip_path = get_zip_path(jar_path)

      file_obj_path = jar_path.replace(zip_path + '/', '')
      with ZipFile(zip_path) as zip_f:
        with zip_f.open(file_obj_path) as zf, open(new_jar_path, 'wb') as f:
          shutil.copyfileobj(zf, f)

    jar_path = new_jar_path

  if not Path(jar_path).exists():
    raise Exception('Could not find JAR path "{}"'.format(jar_path))

  return jar_path

def rs_to_rows(rs):
  "Get Rows from JPype resultset"
  Row = None
  rows = []
  while rs.next():
    cols = [c.name for c in rs.columns]
    if not Row:
      Row = namedtuple('Row', cols)
    rows.append(Row(*rs.currentRow))
  return rows



# Parallelism Lib
'''
Threads
Multiprocessing
'''
import os, socket, time
from .helpers import (log, slog, elog, get_exception_message, now,
                      ndelta_seconds, get_home_path, cleanup_pid, register_pid,
                      get_pid_path)
from .diskio import (write_file, read_file)
from functools import wraps

from collections import (
  OrderedDict,
  namedtuple,
)
from threading import (
  Thread,
  Lock,
)
from multiprocessing import Process
from multiprocessing import (Queue)

all_threads = OrderedDict()
all_processes = OrderedDict()
th_lock = Lock()


def pipe_post(pipe, obj):
  "Send object through multiprocessing Pipe and wait for response"


class Pipe:
  def __init__(self):
    import multiprocessing
    self.lock = multiprocessing.Lock()
    self.parent, self.child = multiprocessing.Pipe()

  def emit_to_parent(self, obj):
    "Send object through multiprocessing Pipe and wait for response"
    self.child.send(obj)
    return self.child.recv()

  def emit_to_child(self, obj):
    "Send object through multiprocessing Pipe and wait for response"
    self.parent.send(obj)
    return self.parent.recv()

  def send_to_parent(self, obj):
    "Send object through multiprocessing Pipe"
    self.child.send(obj)

  def send_to_child(self, obj):
    "Send object through multiprocessing Pipe"
    self.parent.send(obj)

  def recv_from_parent(self, timeout=-1):
    "Receive object through multiprocessing Pipe. timeout in sec. timeout=-1 means wait forever."
    s_t = now()
    if timeout == -1:
      return self.child.recv()
    elif timeout == 0:
      if self.child.poll():
        return self.child.recv()
    else:
      while 1:
        if self.child.poll():
          return self.child.recv()
        elif ndelta_seconds(s_t) <= timeout:
          break
      return None

  def recv_from_child(self, timeout=-1):
    "Receive object through multiprocessing Pipe. timeout in sec. timeout=-1 means wait forever."
    s_t = now()
    if timeout == -1:
      return self.parent.recv()
    elif timeout == 0:
      if self.parent.poll():
        return self.parent.recv()
    else:
      while 1:
        if self.parent.poll():
          return self.parent.recv()
        elif ndelta_seconds(s_t) <= timeout:
          break
      return None


class Worker:
  kwargs_reserved = ['worker']

  def __init__(self,
               name,
               type,
               fn,
               log=log,
               args=[],
               kwargs={},
               start=False,
               kill_if_running=False,
               pid_folder=None):
    self.hostname = socket.gethostname()
    self.name = name
    self.type = type
    self.pipe = Pipe()  # 2-way
    self.child_q = Queue()
    self.parent_q = Queue()
    self.lock = Lock()
    self.fn = fn
    self.log = log
    self.args = args
    self.kwargs = kwargs
    self.kill_if_running = kill_if_running
    self.started = False
    pid_folder = pid_folder if pid_folder else get_home_path()
    self.pid_file = get_pid_path(name, pid_folder)
    self.status = None

    for key in self.kwargs_reserved:
      if key in kwargs:
        log('kwargs = {}'.format(kwargs))
        raise Exception(
          'Cannot use reserved word "{}" in kwargs for worker {}'.format(
            key, name))

    self.kwargs['worker'] = self
    self.process = Process(target=fn, args=self.args, kwargs=self.kwargs)
    self.process.name = name

    if start:
      self.start()

  def start(self):
    # kill existing PID
    cleanup_pid(self.pid_file, kill_if_running=self.kill_if_running)

    self.process.start()
    self.started = now()
    self.pid = self.process.pid
    register_pid(self.pid_file, self.pid, clean_atexit=True)

  def stop(self):
    self.process.terminate()

    # delete pid file
    cleanup_pid(self.pid_file, kill_if_running=True)

  def put_parent_q(self, obj):
    self.parent_q.put(obj)

  def put_child_q(self, obj):
    self.child_q.put(obj)

  def get_parent_q(self):
    if not self.parent_q.empty():
      return self.parent_q.get()
    else:
      return None

  def get_child_q(self):
    if not self.child_q.empty():
      return self.child_q.get()
    else:
      return None


def kill_processes(name):
  import sys, psutil, signal

  signal_val = int(sys.argv[2]) if len(sys.argv) == 3 else signal.SIGTERM
  signal_val = 9

  user = os.getenv('USER')
  pid = os.getpid()

  killed = []
  for proc in psutil.process_iter():
    if proc.username() != user: continue
    cmdline = ' '.join(proc.cmdline()).replace('\n', ' ')
    if not name in cmdline: continue
    if proc.pid == pid: continue
    print("Killing {} : {}".format(proc.pid, cmdline))
    killed.append(proc.pid)
    os.kill(proc.pid, signal_val)  # else, kill


def interrupt(proc_name=None, th_name=None):
  global all_threads, all_processes, th_lock

  def kill_proc(p):
    log('Killing process PID {} -> {}'.format(p.pid, p.name))
    p.terminate()
    os.system('kill -9 {}'.format(p.pid))

  def kill_thread(t):
    log('Killing thread -> {}'.format(t.name))
    t.cancel()

  if proc_name:
    kill_proc(all_processes[proc_name])
    return

  if th_name:
    kill_thread(all_threads[th_name])
    return

  for proc in all_processes.values():
    with th_lock:
      kill_proc(proc)

  for th in all_threads.values():
    with th_lock:
      kill_thread(th)


def get_running_threads():
  global all_threads
  threads_running = [th.name for th in all_threads.values() if th.isAlive()]
  return threads_running


def get_running_processes():
  global all_processes
  proc_running = [
    proc.name for proc in all_processes.values() if proc.is_alive()
  ]
  return proc_running


def join_threads():
  "Join all running threads: wait until done"
  global all_threads
  running = [1, 1]
  while running:
    time.sleep(1)
    with th_lock:
      running = [th for th in all_threads.values() if th.isAlive()]
    # th.join()


def run_async(func):
  """
    run_async(func)
      function decorator, intended to make "func" run in a separate
      thread (asynchronously).
      Returns the created Thread object
      E.g.:
      @run_async
      def task1():
        do_something
      @run_async
      def task2():
        do_something_too
      t1 = task1()
      t2 = task2()
      ...
      t1.join()
      t2.join()
  """

  @wraps(func)
  def async_func(*args, **kwargs):
    global all_threads, th_lock

    f_name = '_'.join([func.__name__, str(args), str(kwargs)])
    func_hl = None

    if f_name in all_threads:
      if all_threads[f_name].isAlive():
        func_hl = all_threads[f_name]

    if not func_hl:
      func_hl = Thread(name=f_name, target=func, args=args, kwargs=kwargs)
      func_hl.start()
      log('Thread Started -> {}'.format(func_hl.name))

    with th_lock:
      all_threads[f_name] = func_hl

    return func_hl

  return async_func


def run_async_process(func):
  """
    run_async_process(func)
      function decorator, intended to make "func" run in a separate
      thread (asynchronously).
      Returns the created Thread object
      E.g.:
      @run_async
      def task1():
        do_something
      @run_async
      def task2():
        do_something_too
      t1 = task1()
      t2 = task2()
      ...
      t1.join()
      t2.join()
  """
  from multiprocessing import Process

  @wraps(func)
  def async_func(*args, **kwargs):
    global all_processes, th_lock

    f_name = '_'.join([func.__name__, str(args), str(kwargs)])
    func_hl = None

    if f_name in all_processes:
      if all_processes[f_name].is_alive():
        func_hl = all_processes[f_name]

    if not func_hl:
      func_hl = Process(name=f_name, target=func, args=args, kwargs=kwargs)
      func_hl.start()
      log('Process Started PID {} -> {}'.format(func_hl.pid, func_hl.name))

    with th_lock:
      all_processes[f_name] = func_hl

    return func_hl

  return async_func


# Web Lib
'''
HTML
eMailing
Telegram
Web Hooks
'''
import os, socket, json
import re

## HTML ####################

from xutil.helpers import (slog, elog, log, get_kw)
from xutil.parallelism import Pipe, Worker
from flask import Flask

# from scrapy.selector import Selector
# from scrapy.http import HtmlResponse


def generate_rmd_html(rmd_file,
                      output_file='NULL',
                      output_dir='NULL',
                      var_data={},
                      print_output=False,
                      serve=False):
  "rmd_file is the file path to the Rmd file. HTML document will be generated in same folder as Rmd file."
  import os, subprocess
  pandoc_path = os.getenv('R_PANDOC_HOME')
  os.environ["PATH"] = '{}:{}'.format(pandoc_path, os.environ[
    "PATH"]) if pandoc_path not in os.environ["PATH"] else os.environ["PATH"]

  # Apply key-value variables to Rmd file
  if len(var_data):
    if output_file == 'NULL':
      raise Exception(
        'Parameter "output_file" must be specified when providing "var_data" values!'
      )
    if output_dir == 'NULL':
      raise Exception(
        'Parameter "output_dir" must be specified when providing "var_data" values!'
      )
    if not Path(rmd_file).exists():
      raise Exception('File "{}" does not exists'.format(rmd_file))

    rmd_new = output_dir + '/' + output_file.replace('.html', '') + '.Rmd'
    output_file = output_file if '.' in output_file else output_file + '.html'

    if rmd_new == rmd_file:
      raise Exception(
        'Parameter name of "output_file" must be different from  "rmd_file"!')

    os.system("cp -f {} {}".format(rmd_file, rmd_new))  # backup copy
    rmd_text = read_file(rmd_new)
    rmd_text = str_format(rmd_text, var_data, '$')
    write_to_file(rmd_new, rmd_text)
    rmd_file = rmd_new

  add_quotes = lambda t: '"' + t + '"' if t != 'NULL' else t
  cmd = """R -e 'library(rmarkdown); rmarkdown::render("{rmd_file}", "html_document", {output_file}, {output_dir})'""".format(
    rmd_file=rmd_file,
    output_file=add_quotes(output_file),
    output_dir=add_quotes(output_dir),
  )
  out_text = subprocess.getoutput(cmd)

  success = 'Output created: ' in out_text
  if print_output or not success: print(out_text)

  html_file = [
    line for line in out_text.splitlines() if 'Output created: ' in line
  ][0].split(': ')[-1] if success else None

  if serve: serve_http_root(port=5567, http_path=html_file)

  return html_file


def extract_text_from_html(html_source):
  from bs4 import BeautifulSoup

  def visible(element):
    if element.parent.name in [
        'style', 'script', '[document]', 'head', 'title'
    ]:
      return False
    elif re.match('<!--.*-->', str(element)):
      return False
    return True

  soup = BeautifulSoup(html_source, "lxml")
  texts = soup.findAll(text=True)
  text = "\r\n".join(filter(visible, texts))
  return text

def css_select(body, css_string):
  from scrapy.selector import Selector
  return Selector(text=body).css(css_string)

def css_value(row, css_string):
   return row.css(css_string)[0].extract().strip() if row.css(css_string) else None


## EMail ################


def send_email_html(smtp,
                    email_user,
                    email_pwd,
                    to_address,
                    subject,
                    body_text,
                    images_jpg_path=[]):
  from email.mime.multipart import MIMEMultipart
  from email.mime.text import MIMEText
  from email.mime.image import MIMEImage

  # Define these once; use them twice!
  strFrom = email_user
  strTo = to_address  #must be a list

  # Create the root message and fill in the from, to, and subject headers
  msgRoot = MIMEMultipart('related')
  msgRoot['Subject'] = subject
  msgRoot['From'] = strFrom
  msgRoot['To'] = strTo
  msgRoot.preamble = 'This is a multi-part message in MIME format.'

  # Encapsulate the plain and HTML versions of the message body in an
  # 'alternative' part, so message agents can decide which they want to display.
  msgAlternative = MIMEMultipart('alternative')
  msgRoot.attach(msgAlternative)

  msgText = MIMEText(extract_text_from_html(body_text))
  msgAlternative.attach(msgText)

  # We reference the image in the IMG SRC attribute by the ID we give it below
  msgText = MIMEText(body_text, 'html')
  msgAlternative.attach(msgText)

  #Upload Images
  i = 0
  for img_path in images_jpg_path:
    fp = open(img_path, 'rb')
    msgImage = MIMEImage(fp.read())
    fp.close()
    i += 1
    msgImage.add_header('Content-ID', '<image' + str(i) + '>')
    msgRoot.attach(msgImage)

  # Send the email (this example assumes SMTP authentication is required)
  smtp.login(email_user, email_pwd)
  smtp.sendmail(strFrom, strTo, msgRoot.as_string())
  smtp.quit()


def send_from_gmail(to_address, subject, body_text, images_jpg_path=[]):
  import smtplib
  smtp = smtplib.SMTP("smtp.gmail.com", 587, 'localhost', 30)
  email_user = os.getenv("SMTP_USER")
  email_pwd = os.getenv("SMTP_PASSWORD")
  smtp.ehlo()
  smtp.starttls()
  send_email_html(
    smtp=smtp,
    email_user=email_user,
    email_pwd=email_pwd,
    to_address=to_address,
    subject=subject,
    body_text=body_text,
    images_jpg_path=images_jpg_path)


def send_from_outlook(to_address, subject, body_text, images_jpg_path=[]):
  import smtplib
  smtp = smtplib.SMTP("smtp-mail.outlook.com", 587, 'localhost', 30)
  email_user = os.getenv("SMTP_USER")
  email_pwd = os.getenv("SMTP_PASSWORD")
  smtp.ehlo()
  smtp.starttls()
  send_email_html(
    smtp=smtp,
    email_user=email_user,
    email_pwd=email_pwd,
    to_address=to_address,
    subject=subject,
    body_text=body_text,
    images_jpg_path=images_jpg_path)


def send_from_yahoo(to_address, subject, body_text, images_jpg_path=[]):
  import smtplib
  smtp = smtplib.SMTP("smtp.mail.yahoo.com", 465, 'localhost', 30)
  email_user = os.getenv("SMTP_USER")
  email_pwd = os.getenv("SMTP_PASSWORD")
  smtp.ehlo()
  smtp.starttls()
  send_email_html(
    smtp=smtp,
    email_user=email_user,
    email_pwd=email_pwd,
    to_address=to_address,
    subject=subject,
    body_text=body_text,
    images_jpg_path=images_jpg_path)


def send_email_exchange(to_address,
                        subject,
                        body_text,
                        sender=None,
                        attachments=[],
                        image_paths=[],
                        html=False):
  import smtplib
  from os.path import basename
  from email.mime.application import MIMEApplication
  from email.mime.multipart import MIMEMultipart
  from email.mime.text import MIMEText
  from email.utils import COMMASPACE, formatdate

  msg = MIMEMultipart('related') if html else MIMEMultipart()
  sender = sender if sender else os.getenv("SMTP_USER")
  to_address = to_address if isinstance(to_address, list) else [to_address]

  msg['From'] = sender
  msg['To'] = ','.join(to_address)
  msg['Subject'] = subject

  if html:
    msg.preamble = 'This is a multi-part message in MIME format.'
    msgAlternative = MIMEMultipart('alternative')
    msg.attach(msgAlternative)

    msgText = MIMEText(extract_text_from_html(body_text))
    msgAlternative.attach(msgText)

    msgText = MIMEText(body_text, 'html')
    msgAlternative.attach(msgText)

  else:
    msg.attach(MIMEText(body_text, 'plain'))

  for i, img_path in enumerate(image_paths):
    with open(img_path, 'rb') as fp:
      msgImage = MIMEImage(fp.read())
    msgImage.add_header('Content-ID', '<image' + str(i + 1) + '>')
    msg.attach(msgImage)

  for f in attachments:
    with open(f, "rb") as file:
      part = MIMEApplication(file.read(), Name=basename(f))
    part['Content-Disposition'] = 'attachment; filename="%s"' % basename(f)
    msg.attach(part)

  # Send the message via our SMTP server
  SMTP_SERVER = os.getenv("SMTP_SERVER")
  if not SMTP_SERVER:
    raise Exception('Env SMTP_SERVER is not defined!')

  s = smtplib.SMTP(SMTP_SERVER)
  s.sendmail(sender, to_address, msg.as_string())
  s.quit()
  log('Sent Email "{}" succesfully!'.format(subject))


def get_imap_messages(mailbox='inbox',
                      recent_limit=100,
                      subject_filter=None,
                      body_filter=None,
                      unread_only=False,
                      from_filter=None,
                      to_filter=None,
                      search_items=['ALL']):
  """
  search_items:
    ['BODY', '"HELLO"']
    ['FROM', '"HELLO"']
    ['SUBJECT', '"HELLO"']
    ['UNSEEN']
    See https://tools.ietf.org/html/rfc3501#section-6.4.4
  """
  import email, imaplib
  USERNAME = os.getenv('IMAP_USERNAME')
  PASSWD = os.getenv('IMAP_PASSWD')
  SMTP_SERVER = os.getenv('IMAP_SMTP_SERVER')
  SMTP_PORT = 993

  if not USERNAME:
    raise Exception('Environement Var "IMAP_USERNAME" is not set!')
  if not PASSWD:
    raise Exception('Environement Var "IMAP_PASSWD" is not set!')
  if not SMTP_SERVER:
    raise Exception('Environement Var "IMAP_SMTP_SERVER" is not set!')

  if subject_filter:
    search_items += ['SUBJECT', '"{}"'.format(subject_filter)]

  if body_filter:
    search_items += ['BODY', '"{}"'.format(body_filter)]

  if from_filter:
    search_items += ['FROM', '"{}"'.format(from_filter)]

  if to_filter:
    search_items += ['TO', '"{}"'.format(to_filter)]

  if unread_only:
    search_items += ['UNSEEN']

  mail = imaplib.IMAP4_SSL(SMTP_SERVER)
  mail.login(USERNAME, PASSWD)
  mail.select(mailbox)

  resp, data = mail.search(None, *search_items)
  ids = data[0].split()
  ids = ids[-recent_limit:] if len(ids) > recent_limit else ids
  messages = []

  for i in ids:
    resp, data = mail.fetch(i, '(RFC822)')

    message = email.message_from_bytes(data[0][1])
    body = message.get_payload()
    body = '{} chars'.format(len(body)) if len(body) > 200 else body

    message_data = {
      'date': message['Date'],
      'from': message['From'],
      'to': message['To'],
      'subject': message['Subject'],
      'body': body,
    }
    messages.append(message_data)

    # summary = '''
    # Date: {date}
    # From: {from}
    # To: {to}
    # Subject: {subject}
    # Body: {body}
    # '''.format(**message_data)
    # print(summary)

  return messages


## Web Hooks & API ################


def tk(app_name, action, name, value):
  "Track app actions for analysis"
  params = dict(
    AppName=app_name,
    Action=action,
    Name=name,
    Value=value,
  )

  URL = os.getenv('APP_TRACKER_URL')
  if not URL:
    raise (Exception("Env Var APP_TRACKER_URL is not defined"))

  try:
    requests.post(URL, params=params)
  except:
    pass


def get_cookie_session_id(req):
  import random, string
  session_id = (req.cookies.get(cookie_session_key) or ''.join(
    random.SystemRandom().choice(string.ascii_uppercase + string.digits +
                                 string.ascii_lowercase) for _ in range(48)))
  return session_id


def process_request(req):
  val_dict = req.values.to_dict()
  form_dict = req.form
  data_dict = json.loads(req.data.decode('utf-8')) if req.data else {}
  return (val_dict, form_dict, data_dict)


def create_sio_client(host, port, response_map={}):
  """
  A Socket IO client
  https://pypi.org/project/socketIO-client/

  Example: 
  on_connect = lambda *args: print('connect')
  aaa_response = lambda *args: print('on_aaa_response', args)

  response_map = {
    'on_connect': on_connect,
    'aaa_response': aaa_response,
  }
  """
  from socketIO_client import SocketIO, LoggingNamespace
  sio_client = SocketIO(host, port, LoggingNamespace)
  for key in response_map:
    sio_client.on(key, response_map[key])

  return sio_client


class MyFlask(Flask):
  jinja_options = Flask.jinja_options.copy()
  jinja_options.update(
    dict(
      block_start_string='(%',
      block_end_string='%)',
      variable_start_string='((',
      variable_end_string='))',
      comment_start_string='(#',
      comment_end_string='#)',
    ))


class WebApp:
  '''
  A boilerplate for a web application using evenlet with flask & socketio.

  from xutil import WebApp, process_request
  app = WebApp(name='App1')

  @app.route('/')
  def index():
    """Serve the client-side application."""
    (val_dict, form_dict, data_dict) = process_request(request)
    app.log('Requested "/"')
    return 'Hi!'

  @app.on('connect')
  def connect(sid, environ):
    app.log('connect ' + sid)

  @app.on('message')
  def message(sid, data):
    app.log('message ' + str(data))

  @app.on('disconnect')
  def disconnect(sid):
    app.log('disconnect ' + sid)

  app.run(port=5899)
  '''

  def __init__(self, name, root_path=None, flask_app=None):
    import os, sys, time, json
    import socketio, socket
    from flask import Flask, request, make_response, render_template, send_from_directory, redirect, url_for

    self.name = name
    self.cookie_session_key = name + '_SID'
    self.sio = socketio.Server()
    self.flask_app = flask_app if flask_app else MyFlask(
      __name__, root_path=root_path)
    self.log = log
    self.request = request
    self.make_response = make_response
    self.redirect = redirect
    self.url_for = url_for
    self.render_template = render_template
    self.send_from_directory = send_from_directory

    if os.getenv('SECURE_SSL_REDIRECT', default=False):
      from flask_sslify import SSLify
      sslify = SSLify(self.flask_app)

    # Wrapper functions
    self.route = self.flask_app.route
    self.on = self.sio.on
    self.emit = self.sio.emit

  def run(self, port, host='0.0.0.0', debug=False, url_suffix='', **kwargs):
    import eventlet, socketio
    import eventlet.wsgi

    if 'worker' in kwargs:
      self.worker: Worker = kwargs['worker']
      self.pipe: Pipe = self.worker.pipe
      self.log = self.worker.log
    if 'pipe' in kwargs:
      self.pipe: Pipe = kwargs['pipe']
    self.log = get_kw('log', self.log, kwargs)

    self.port = int(port)
    hostname = socket.gethostname() if host == '0.0.0.0' else host
    self.base_url = 'http://{}:{}'.format(hostname, self.port)

    # remember to use DEBUG mode for templates auto reload
    # https://github.com/lepture/python-livereload/issues/144
    self.flask_app.debug = debug

    app = socketio.Middleware(self.sio, self.flask_app)

    log('*Web Server PID is {}'.format(os.getpid()))
    log("*URL -> " + self.base_url + url_suffix)

    eventlet.wsgi.server(eventlet.listen((host, self.port)), app)

  def proc_request(self):
    return process_request(self.request)

  def parse_sio_cookies(self, sio_environ):
    "Extract Cookies from SIO Env data"
    cookies = {}
    headers = sio_environ.get('headers_raw', None)
    cookies_raw = [h[1] for h in headers if h[0] == 'Cookie']
    if not (headers and cookies_raw):
      log('~~Headers and/or Cookies not found.')
      return cookies

    cookies_raw = cookies_raw[0]
    cookies_kv = [kv.strip().split('=') for kv in cookies_raw.split(';')]
    for kv in cookies_kv:
      kv = [kv[0], '='.join(kv[1:])]
      key, val = kv
      cookies[key] = val

    return cookies

  def get_cookie_session_id(self):
    import random, string
    session_id = (self.request.cookies.get(self.cookie_session_key) or ''.join(
      random.SystemRandom().choice(string.ascii_uppercase + string.digits +
                                   string.ascii_lowercase) for _ in range(48)))
    return session_id


from .base import DBConn
import sys, os
from xutil.helpers import (get_exception_message, now, log, struct, log)
from xutil.diskio import read_yaml, read_csvD, write_jsonl, read_jsonl


class HiveConn(DBConn):
  "Hive Connection"

  data_map = dict(
    string='VARCHAR()',
    integer='NUMBER()',
    decimal='NUMBER()',
    date='DATE',
    datetime='DATE',
    text='CLOB',
  )

  reverse_data_map = dict(
    INT_TYPE='integer',
    TIMESTAMP_TYPE='datetime',
    STRING_TYPE='string',
  )

  def connect(self):
    "Connect / Re-Connect to Database"
    c = self._cred

    # import pyhs2
    # self.connection = pyhs2.connect(
    #   host=c.host,
    #   port=c.port,
    #   authMechanism="KERBEROS",
    #   password=c.password,
    #   user=c.user)
    # self.connection.autocommit = True

    from impala.dbapi import connect
    self.connection = connect(
      host=c.host,
      port=c.port,
      **c.kwargs)
    self.cursor = None

    self.name = "Hive"
    self.username = c.user

  def set_variables(self):

    self.col_name_id = 0

    self.error_msg = dict(table_not_exist='table or view does not exist', )

    self.sql_template = dict(
      drop_table='drop table {} cascade constraints',
      insert="INSERT {options} INTO {table} ({names}) VALUES ({values})",
      create_table="create table if not exists {table} ({col_types}) {options}",
      create_index="create index {index} on {table} ({cols})",
      analyze='''select count(*) as tot_cnt, {fields} from {table}''',
      columns='''describe {schema}.{table}''',
      tables='''show tables in {schema}''',
      schemas='''show databases''',
      ddl='''show create table {schema}.{table}''',
      insert_option='',
    )

    self.temp_path = self.tmp_folder

  def get_engine(self, echo=False):
    "Could implement PyHive?"
    return None


  def get_cursor_metadata(self, schema, table, sample_row=None):
    "Get current cursor column metadata"
    samples = sample_row._asdict() if sample_row else {
      fix_field(f['columnName']): None
      for f in self.cursor.getSchema()
    }
    data = [
      self.nt_cursor_metadata(
        schema,
        table,
        fix_field(f['columnName']),
        f['type'],
        n + 1,
        samples[fix_field(f['columnName'])],
        schema + '.' + table,
        schema + '.' + table + '.' + fix_field(f['columnName']),
      ) for n, f in enumerate(self.cursor.getSchema())
    ]

    return data

  def _do_execute(self, sql, cursor=None):
    super(HiveConn, self)._do_execute(sql, cursor=cursor)
    cleanse_f = lambda f: f.split('.')[-1] if '.' in f else f
    self._fields = [cleanse_f(f) for f in self._fields]


  # def get_cursor_fields(self, cursor=None):
  #   "Get fields of active Select cursor"
  #   cursor = cursor if cursor else self.cursor
  #   fields = [fix_field(f['columnName']) for f in cursor.getSchema()]
  #   return fields


  # def select(self,
  #            sql,
  #            nt_name='Record',
  #            dtype='namedtuple',
  #            limit=None,
  #            date_fields_=[],
  #            echo=True,
  #            date_conv=True,
  #            date_conv_str=False,
  #            log=log):
  #   "Select from SQL, return list of namedtuples"
  #   data = super(HiveConn, self).select(
  #     sql, nt_name, dtype='dataframe', echo=echo)

  #   # Hive returns date fields as text. Not useful.
  #   # This is to convert from text to date using ever-efficient dataframes
  #   fields = [f['columnName'] for f in self.cursor.getSchema()]

  #   date_fields = [fix_field(f['columnName']) for f in self.cursor.getSchema() \
  #                  if 'DATE' in f['type'] or 'TIME' in f['type']]

  #   if date_conv_str:
  #     date_fields = [fix_field(f['columnName']) for f in self.cursor.getSchema() \
  #                    if 'DATE' in f['type'] or 'TIME' in f['type'] or 'date' in f['columnName'].lower() \
  #                    or f['columnName'].lower().endswith('_dt')]

  #   date_fields = date_fields + [f for f in date_fields_ if f in fields]

  #   if date_fields and date_conv:
  #     if echo: log('>> Attempting to convert date strings to date types.')
  #     str_io = StringIO()  # text in memory
  #     data.to_csv(str_io, index=False)
  #     str_io.seek(0)  # go to beginning of virtual file
  #     try:
  #       data = read_csvD(str_io, date_cols=list(set(date_fields)))
  #       if echo: log('>> Done Converting')
  #     except Exception as e:
  #       message = get_exception_message().lower()
  #       log('ERROR Converting: ' + get_exception_message())

  #   if dtype == 'namedtuple':
  #     data = df_to_list(data)

  #   return data

  @staticmethod
  def execute_cli(sql,
                  output_path=None,
                  delimiter=',',
                  ret_data=True,
                  date_cols=[],
                  use_beeline=False,
                  ret_df=False,
                  echo=True):
    "Execute using command line"
    deli = '\\' + delimiter if delimiter in ['|'] else delimiter

    cmd_tmpl = self.variables[
      'beeline_cmd'] if use_beeline else self.variables['hive_cmd']

    if output_path: cmd_tmpl = '{} > {}'.format(cmd_tmpl, output_path)
    command = cmd_tmpl.format(sql=sql, deli=deli)
    # print(command)
    s_t = datetime.datetime.now()
    process = subprocess.Popen(
      command,
      shell=True,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
    )
    # print(command)

    pid = process.pid

    if echo: log("Running Hive query. PID is {}".format(pid))

    data = []
    Rec = None

    try:
      if not output_path and ret_data:
        v_file_text = b''.join(process.stdout).decode('utf8')
        # print(v_file_text)
        v_file = StringIO(v_file_text)
        reader = csv.reader(v_file, delimiter=',')
        Rec = namedtuple('Rec', [
          h.replace(' ', '_').split('.')[1] if '.' in h else h
          for h in [str(v) for v in next(reader)]
        ])
        data = [row for i, row in enumerate(imap(Rec._make, reader))]

      process.wait()
      stderr = [line.rstrip().decode('utf8') for line in process.stderr]
      # print(stderr)
      if 'Error: ' in '\n'.join(stderr[-2:]) and use_beeline:
        print('\n'.join(stderr[-2:]))
        print(stderr)
        return []

    except KeyboardInterrupt:
      process.terminate()
      log("PID {} terminated!".format(pid))
      return []

    counter = len(data)
    if output_path:
      if ret_data:
        data = read_csvD(
          output_path, delimiter=delimiter, echo=False, date_cols=date_cols)
        if ret_df:
          fix_col = lambda c: c.replace(' ', '_').split('.')[-1] if '.' in c else c.replace(' ', '_')
          data.rename(
            columns={c: fix_col(c)
                     for c in data.columns}, inplace=True)
        else:
          data = df_to_list(data)

        counter = len(data)
      else:
        if stderr and 'rows selected' in stderr[-3] and use_beeline:
          counter = int(stderr[-3].split()[0].replace(',', ''))
        elif stderr and 'Fetched' in stderr[-1]:
          counter = int(stderr[-1].split()[-2].replace(',', ''))

    output, error = process.communicate()
    returncode = process.returncode

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)

    if echo:
      log("Hive query finished! Got {} records in {} mins [{} r/s].".format(
        counter, mins, rate))

    return data

  def _std_get_columns(self, schema, table, rows):
    "Standardize for get_columns"
    make_rec = lambda r, i: dict(name=r.col_name, type=r.data_type, schema=schema, table=table, nullable=None, default=None, autoincrement=None, id=i)
    data = [make_rec(r, i) for i, r in enumerate(rows)]
    return data

  def _std_get_tables(self, schema, rows):
    "Standardize for get_tables"
    data = [r.tableName for r in rows if not r.isTemporary]
    return data


class Beeline(HiveConn):
  "Any JDBC Connection String"

  cmd_tmpl = """{cmd} > {path}"""

  def get_columns(self, table, names_only=False, echo=False):
    "Get column metadata for table"
    owner, table = table.split('.') if '.' in table else (self.username, table)
    sql = self.sql_template['columns'].format(
      owner=owner,
      table=table,
    )
    Rec = namedtuple('Rec', 'column_name data_type')
    if names_only:
      return [r[0] for r in self.select(sql, table + ' columns', echo=echo)]
    else:
      return [
        Rec(r[0], r[1])
        for r in self.select(sql, table + ' columns', echo=echo)
      ]

  def convert_to_parquet(self, table, order_by=[]):
    "Convert table to parquet format"
    sql = '''use {db};
    drop table if exists {table}z;
    create table {table}z stored as parquet as
    select * from {table} {ord};
    drop table {table};
    '''.format(
      db=table.split('.')[0],
      table=table,
      ord='order by ' + ', '.join(order_by) if order_by else '',
    )

    # get table path
    data = self.select('show create table ' + table, echo=False)
    for i, r in enumerate(data):
      if data[i][0].strip() == 'LOCATION':
        table_hdfs_path = data[i + 1][0].strip().replace("'", '')
        log('HDFS Path: ' + table_hdfs_path)
        break

    # delete z folder, create z temp table
    os.system('hdfs dfs -rm -r -f -skipTrash {}z'.format(table_hdfs_path))
    self.execute(sql, echo=False)

    # delete orignal folder, create final table
    # When renaming, error keeps occuring
    os.system('hdfs dfs -rm -r -f -skipTrash {}'.format(table_hdfs_path))
    sql = '''use {db}; alter table {table}z rename to {table}'''.format(
      db=table.split('.')[0],
      table=table,
    )
    self.execute(sql, echo=False)

  def select_to_csv(self,
                    sql,
                    file_path=None,
                    nt_name='Record',
                    echo=True,
                    echo2=True):
    "Select from SQL, write to csv"

    t_suff = datetime.datetime.now().strftime('%Y%m%d-%H%M')
    pwd_path = '{}/p_{}_{}'.format(self.temp_path, self._cred.name, t_suff)
    file_path = '{}/data_{}.csv'.format(self.temp_path,
                                        t_suff) if not file_path else file_path

    os.system('mkdir -p ' + self.temp_path)
    # write_to_file(self._cred.password, pwd_path)

    command = self.cmd_tmpl.format(
      cmd=self.variables['beeline_cmd'].format(sql),
      path=file_path,
    )

    s_t = datetime.datetime.now()
    process = subprocess.Popen(
      command,
      shell=True,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
    )

    pid = process.pid

    if echo: log("Running Beeline query. PID is {}.".format(pid))

    try:
      process.wait()
      stderr = [line.rstrip().decode('utf8') for line in process.stderr]
      if 'Error: ' in '\n'.join(stderr):
        print('\n'.join(stderr[-2:]))
        raise Exception('Hive Exception: ' + '\n'.join(stderr))

        print(stderr)
        print(command)
        return []

    except KeyboardInterrupt:
      process.terminate()
      log("PID {} terminated!".format(pid))
      return []
    finally:
      os.system('rm -f ' + pwd_path)

    output, error = process.communicate()
    returncode = process.returncode

    counter = int(stderr[-3].split()[0].replace(
      ',', '')) if stderr and 'rows selected' in stderr[-3] else 0

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)

    if echo and echo2:
      log("Wrote {} records to {} in {} mins [{} r/s].".format(
        counter, file_path, mins, rate))

  def execute(self, sql, echo=True, echo2=True):
    "Excecute SQL"
    s_t = datetime.datetime.now()
    t_suff = datetime.datetime.now().strftime('%Y%m%d-%H%M')
    file_path = '{}/data_{}.csv'.format(self.temp_path, t_suff)
    self.select_to_csv(sql, file_path, echo=echo, echo2=False)
    os.system('rm -f ' + file_path)

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)

    if echo:
      log("Beeline query finished in {} secs / {} mins.".format(secs, mins))

  def select(self,
             sql,
             nt_name='Record',
             dtype='namedtuple',
             limit=None,
             date_cols=[],
             echo=True,
             log=log):
    "Select from SQL, return list of namedtuples"
    t_suff = datetime.datetime.now().strftime('%Y%m%d-%H%M')
    file_path = '{}/data_{}.csv'.format(self.temp_path, t_suff)

    # Run Query stdout to file
    s_t = datetime.datetime.now()
    self.select_to_csv(sql, file_path, nt_name=nt_name, echo=echo, echo2=False)

    data = read_csvD(file_path, echo=False, date_cols=date_cols)
    os.system('rm -f ' + file_path)

    if dtype == 'dataframe':
      fix_col = lambda c: c.replace(' ', '_').split('.')[-1] if '.' in c else c.replace(' ', '_')
      data.rename(columns={c: fix_col(c) for c in data.columns}, inplace=True)
    else:
      data = df_to_list(data)

    counter = len(data)
    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)

    if echo:
      log("Beeline SELECT query finished! Got {} records in {} mins [{} r/s].".
          format(counter, mins, rate))

    return data

  def _concat_fields(self, fields, as_text=False):
    if as_text:
      fields = ['cast({} as string)'.format(f) for f in fields]
    return 'concat({})'.format(', '.join(fields))

  def get_tables(self, schema, echo=True):
    "Get metadata for tables."
    Rec = namedtuple('Table', 'schema table')

    def get_rec(table):
      r_dict = dict(schema=schema, table=table)
      return Rec(**r_dict)

    sql_tmpl = self._template('metadata.tables')
    tables = [r[1] for r in self.select(sql_tmpl.format(schema=schema))]
    rows = [get_rec(t) for t in sorted(tables)]
    return rows

  def get_columns(self, obj, object_type=None, echo=False):
    "Get column metadata for table"
    Rec = namedtuple(
      'Columns',
      'schema table column_name type column_order nullable default autoincrement'
    )
    schema, table = self._split_schema_table(obj)

    def get_rec(r_dict, column_order):
      r_dict['schema'] = schema
      r_dict['table'] = table
      r_dict['column_name'] = r_dict['col_name']
      r_dict['type'] = r_dict['data_type']
      r_dict['column_order'] = column_order
      r_dict['nullable'] = None
      r_dict['default'] = None
      r_dict['autoincrement'] = None

      del r_dict['col_name']
      del r_dict['data_type']
      if 'attrs' in r_dict: del r_dict['attrs']
      return Rec(**r_dict)

    sql_tmpl = self._template('metadata.columns')
    rows = self.select(sql_tmpl.format(table=table, schema=schema))
    rows = [get_rec(r.asDict(), i + 1) for i, r in enumerate(rows)]

    return rows


import datetime, csv, time, pandas, os
from xutil.database.base import DBConn
from xutil.helpers import (get_exception_message, now, log, struct,
                           is_gen_func, isnamedtupleinstance, get_dir_path)
from xutil.diskio import read_yaml, read_file, write_jsonl, read_jsonl
from collections import namedtuple

############################################################
"""Experiment to have two threads to write to parquet files:
Thread 1 would be the database thread. All it does it collect records and append them into a buffer list.
Thread 2 would poll that buffer list, once criteria is met (size of list or Thread 1 finished), it will yield the batch dataframe (to be written to parquet) all on the same thread. This way it does not interefere with Thread 1.
So 3 variables would be shared: the field list, the records buffer list and the finished indicator of thread #1.
Below code chunck does not work as intended -- need to seperate the database thread completely.
"""
from threading import (
  Thread,
  Lock,
)

th_lock = Lock()
buf_rows = []
buf_df = []
making_df = False


def make_df(rows, _fields):
  global buf_df, buf_rows, making_df
  buf_rows += rows

  if len(buf_rows) > 100000:
    with th_lock:
      log('making buf_df')
      making_df = True
      buf_df = pandas.DataFrame(buf_rows, columns=_fields)
      making_df = False
      buf_rows = []


def make_batch_df(rows, _fields):
  global buf_df, buf_rows, making_df
  df = []

  if rows and not making_df:
    th = Thread(name='make_df', target=make_df, args=(rows, _fields))
    th.start()

  with th_lock:
    if len(buf_df):
      df = buf_df
      buf_df = []

  return df


############################################################


class PostgreSQLConn(DBConn):
  "PostgreSQL Connection"

  _to_text = lambda self, t: t+'::text'

  data_map = dict(
    string='VARCHAR',
    integer='integer',
    decimal='double',
    date='DATE',
    datetime='TIMESTAMP',
    timestamp='TIMESTAMP',
    text='text',
  )

  # from table pg_catalog.pg_type
  reverse_data_map = {
    16: 'integer',
    # 17:'_bytea',
    # 18:'name',
    18: 'string',
    # 19:'_name',
    20: 'integer',
    # 21:'int2vector',
    21: 'integer',
    # 22:'_int2vector',
    23: 'integer',
    # 24:'_regproc',
    25: 'string',
    # 26:'oidvector',
    # 26:'_oid',
    # 27:'_tid',
    # 28:'_xid',
    # 29:'_cid',
    # 30:'_oidvector',
    114: 'string',
    142: 'string',
    # 600:'lseg',
    # 600:'_point',
    # 600:'box',
    # 601:'_lseg',
    # 602:'_path',
    # 603:'_box',
    # 604:'_polygon',
    # 628:'_line',
    # 650:'_cidr',
    700: 'float',
    # 701:'point',
    701: 'double',
    # 701:'line',
    # 702:'_abstime',
    # 703:'_reltime',
    # 704:'_tinterval',
    # 718:'_circle',
    790: 'number',
    # 829:'_macaddr',
    # 869:'_inet',
    # 1033:'_aclitem',
    1042: 'string',
    1043: 'string',
    1082: 'datetime',
    1083: 'datetime',
    1114: 'datetime',
    1184: 'datetime',
    # 1186:'_interval',
    # 1266:'_timetz',
    # 1560:'_bit',
    # 1562:'_varbit',
    1700: 'number',
    # 1790:'_refcursor',
    # 2202:'_regprocedure',
    # 2203:'_regoper',
    # 2204:'_regoperator',
    # 2205:'_regclass',
    # 2206:'_regtype',
    # 2249:'_record',
    2275: 'string',
    2950: 'string',
    # 2970:'_txid_snapshot',
    # 3220:'_pg_lsn',
    # 3614:'_tsvector',
    # 3615:'_tsquery',
    # 3642:'_gtsvector',
    # 3734:'_regconfig',
    # 3769:'_regdictionary',
    3802: 'string',
  }

  object_type_map = {'TABLE': 'BASE TABLE', 'VIEW': 'VIEW'}

  def set_variables(self):
    self.connection.autocommit = False
    self.connection.execute('commit')

    self.batch_size = 50000

    self.col_name_id = 2

    self.error_msg = dict(table_not_exist='table or view does not exist', )

    self.sql_template = dict(
      drop_table='drop table if exists {}',
      insert="INSERT {options} INTO {table} ({names}) VALUES ({values})",
      create_table="create table {table}\n({col_types})",
      create_index="create index {index} on {table} ({cols})",
      insert_option='',
      schemas=
      'select schema_name from information_schema.schemata order by schema_name',
      tables=
      '''SELECT '{schema}' as schema, table_name as object_name FROM information_schema.tables WHERE UPPER(table_type) = UPPER('{obj_type}') AND UPPER(table_schema) = UPPER('{schema}') order by table_name''',
      columns='''select
      table_schema as owner, table_name, column_name, data_type,
      character_maximum_length as data_length, numeric_precision as data_precision, numeric_scale as data_scale
      from information_schema.columns
      where table_schema = '{schema}'
      and table_name =  '{table}'
      order by ordinal_position''',
      analyze='''select count(*) as tot_cnt, {fields} from {table}''',
      duplicates='''
      with t0 as (
        select
          {fields}, COUNT(1) as c
        from {table}
        group by {fields}
        having count(1) > 1
      )
      select count(1) dup_cnt from t0
      ''',
      replace='''
      INSERT INTO {table} ({names}) values ({values})
      ON CONFLICT ({pK_fields}) DO UPDATE SET {set_fields};
      ''')

  def get_dialect(self, echo=False):
    """SQLAlchemy dialect"""
    from sqlalchemy.dialects import postgresql
    return postgresql

  def create_engine(self, conn_str=None, echo=False):
    import sqlalchemy
    if not conn_str:
      conn_str = ('postgresql://{user}:{passw}@{host}:{port}/{db}?sslmode={sslmode}'.format(
        user=self._cred.user,
        passw=self._cred.password,
        host=self._cred.host,
        port=self._cred.port,
        db=self._cred.database,
        sslmode=self._cred.sslmode,
      ))

    self.engine = sqlalchemy.create_engine(conn_str, echo=echo)

    return self.engine

  # def connect(self):
  #   "Connect / Re-Connect to Database"
  #   import psycopg2
  #   get_conn_str = lambda cred: "dbname='{}' user='{}' host='{}' port={} password='{}' sslmode='{}'".format(
  #     cred.database, cred.user,
  #     cred.host, cred.port,
  #     cred.password, cred.sslmode)

  #   cred = struct(self._cred) if isinstance(self._cred, dict) else None
  #   cred.sslmode = cred.sslmode if 'sslmode' in cred else 'disable'
  #   conn_str = get_conn_str(cred) if cred else self._cred
  #   self.connection = psycopg2.connect(conn_str)
  #   self.cursor = None

  #   self.connection.autocommit = True

  #   cursor = self.get_cursor()

  # def stream(self,
  #            sql,
  #            rec_name='Record',
  #            dtype='namedtuple',
  #            yield_chuncks=False,
  #            chunk_size=None,
  #            limit=None,
  #            echo=True):
  #   "Stream Select from SQL, yield records as they come in"
  #   from psycopg2.extras import NamedTupleCursor
  #   self.reconnect(min_tresh=10)

  #   if echo: log("Streaming SQL for '{}'.".format(rec_name))

  #   autocommit = self.connection.autocommit
  #   if self.connection.autocommit:
  #     self.connection.autocommit = False

  #   self._stream_counter = 0
  #   fetch_size = limit if limit else self.fetch_size
  #   fetch_size = chunk_size if chunk_size else fetch_size
  #   make_rec = None
  #   fields = None
  #   done = False

  #   cursor = self.connection.cursor(
  #     name='cursor_' + str(int(time.time() * 1000)),
  #     cursor_factory=NamedTupleCursor)
  #   cursor.itersize = fetch_size

  #   try:
  #     self._do_execute(sql, cursor)
  #   except Exception as E:
  #     self.connection.rollback()
  #     done = True
  #     raise E

  #   while not done:
  #     if not make_rec:
  #       row = cursor.fetchone()
  #       fields = self.get_cursor_fields(cursor=cursor)
  #       self._fields = fields
  #       rows = [row] if row else []

  #       if dtype == 'tuple':
  #         make_rec = lambda row: list(row)
  #         make_batch = lambda rows: [make_rec(r) for r in rows]
  #       elif dtype == 'dataframe':
  #         yield_chuncks = True
  #         make_rec = lambda row: list(row)
  #         make_batch = lambda rows: pandas.DataFrame([list(r) for r in rows], columns=self._fields)
  #         # make_batch = lambda rows: make_batch_df(rows, fields)

  #       else:
  #         # since we're using NamedTupleCursor
  #         make_rec = lambda row: row
  #         make_batch = lambda rows: [make_rec(r) for r in rows]

  #       rows = rows + cursor.fetchmany(fetch_size)
  #     else:
  #       rows = cursor.fetchmany(fetch_size)

  #     if not fields:
  #       break

  #     if rows:
  #       if yield_chuncks:
  #         batch = make_batch(rows)
  #         self._stream_counter += len(batch)
  #         if len(batch):
  #           yield batch
  #       else:
  #         for row in rows:
  #           self._stream_counter += 1
  #           yield make_rec(row)
  #           if limit and self._stream_counter == limit:
  #             done = True
  #             break
  #     else:
  #       done = True

  #   cursor.close()
  #   self.connection.commit()
  #   # self.connection.autocommit = autocommit

  def insert_csv(self, table, file_path, delimiter=','):
    s_t = datetime.datetime.now()
    counter = sum(1 for line in open(file_path)) - 1
    file = open(file_path, 'r')
    file.readline()
    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()

    try:
      cursor.copy_from(file, table, sep=delimiter)
    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
      counter, table, mins, rate))

  def insert_ignore(self,
                    table,
                    data,
                    pk_fields=None,
                    commit=True,
                    echo=True,
                    temp_table=None):
    return self.replace(
      table=table,
      data=data,
      pk_fields=pk_fields,
      commit=commit,
      echo=echo,
      sql_tmpl='core.insert_ignore',
      temp_table=temp_table)

  def replace(self,
              table,
              data,
              pk_fields=None,
              commit=True,
              echo=True,
              sql_tmpl='core.replace',
              temp_table=None):
    "Upsert data into database"
    s_t = datetime.datetime.now()

    if not len(data):
      return False

    row = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(row) else 'dict'
    fields = row._fields if mode == 'namedtuple' else sorted(row.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    pk_fields_set = set(pk_fields)
    table = table + '_temp' if temp_table else table
    sql = self._template(sql_tmpl).format(
      table=table,
      set_fields=',\n'.join([
        '{f} = %({f})s'.format(f=f) for i, f in enumerate(fields)
        if f not in pk_fields_set
      ]),
      set_fields2=',\n'.join([
        '{f} = t2.{f}'.format(f=f) for i, f in enumerate(fields)
        if f not in pk_fields_set
      ]),
      names=',\n'.join(['{f}'.format(f=f) for f in fields]),
      pk_fields=', '.join(['{f}'.format(f=f) for f in pk_fields_set]),
      pk_fields_equal=' and '.join(
        ['t1.{f} = t2.{f}'.format(f=f) for f in pk_fields_set]),
      values=',\n'.join(['%({f})s'.format(f=f) for f in fields]),
      temp_table=temp_table,
    )

    if temp_table:
      # drop / create temp table
      self.execute('drop table if exists ' + temp_table, echo=false)
      self.execute(
        'create table {} as select * from {} where 1=0'.format(
          temp_table,
          table,
        ),
        echo=false,
      )
      self.insert(temp_table, data, echo=false)
      self.execute(sql, echo=false)
      self.execute('drop table if exists ' + temp_table, echo=false)
      counter = len(data)

    else:
      connection = self.engine.raw_connection()
      connection.autocommit = False
      cursor = connection.cursor()

      try:
        counter = 0
        if is_gen_func(data):
          batch = [row]

          for row in data:
            batch.append(row)
            if len(batch) == self.batch_size:
              cursor.executemany(sql, batch)
              counter += len(batch)
              batch = []

          if len(batch):
            batch = [r._asdict() for r in batch]
            cursor.executemany(sql, batch)
            counter += len(batch)
        else:
          # cursor.bindvars = None
          cursor.executemany(sql, [r._asdict() for r in data])
          counter += len(data)

        if commit:
          connection.commit()
        else:
          return counter

      except Exception as e:
        log(Exception('Error for SQL: ' + sql))
        raise e

      finally:
        cursor.close()
        connection.commit()
        connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter, table, mins, rate))
    return counter

  def insert(self, table, data, echo=True):
    headers = next(data) if is_gen_func(data) else data[0]

    mode = 'namedtuple' if isnamedtupleinstance(headers) else 'dict'
    fields = headers._fields if mode == 'namedtuple' else sorted(
      headers.keys())
    values = [i + 1
              for i in range(len(fields))] if mode == 'namedtuple' else fields

    connection = self.engine.raw_connection()
    connection.autocommit = False
    cursor = connection.cursor()
    sql = self._template('core.insert').format(
      table=table,
      names=', \n'.join([self._fix_f_name(f) for f in fields]),
      values=', \n'.join(['%s'] * len(values)),
    )

    i = 1

    def get_batch():
      for r, row in enumerate(data):
        yield row

    cols_str = ', '.join(fields)
    copy_sql = '''COPY {} ({}) FROM stdin WITH CSV DELIMITER '|' QUOTE '"' ESCAPE '"' '''.format(table, cols_str)

    try:
      deli = '|'
      s_t = datetime.datetime.now()
      if is_gen_func(data):
        cursor.execute(sql, headers)  # insert first row
        counter = 1
      else:
        counter = 0

      temp_file_path = '{}/batch_sql.csv'.format(self.tmp_folder)
      batch_f = open(temp_file_path, 'w')
      batch_w = csv.writer(batch_f, delimiter=deli, quoting=csv.QUOTE_MINIMAL)

      for r, row in enumerate(data):
        batch_w.writerow(row)
        counter += 1
        # if counter % self.batch_size == 0:
        #   cursor.copy_expert(copy_sql, open(temp_file_path, 'r'))

        #   batch_f = open(temp_file_path, 'w')
        #   batch_w = csv.writer(
        #     batch_f, delimiter=deli, quoting=csv.QUOTE_MINIMAL)

      batch_f.close()
      cursor.copy_expert(copy_sql, open(temp_file_path, 'r'))
      os.remove(temp_file_path)

    except Exception as e:
      log(Exception('Error for SQL: ' + sql))
      raise e
    
    finally:
      cursor.close()
      connection.commit()
      connection.close()

    secs = (datetime.datetime.now() - s_t).total_seconds()
    mins = round(secs / 60, 1)
    rate = round(counter / secs, 1)
    if echo:
      log("Inserted {} records into table '{}' in {} mins [{} r/s].".format(
        counter,
        table,
        mins,
        rate,
      ))
    return counter

  def _concat_fields(self, fields, as_text=False):
    if as_text:
      fields = [self._to_text(f) for f in fields]
    return ' || '.join(fields)



# Network related lib
'''
Samba
SSH
HDFS
'''
import datetime
import socket
import sys
from xutil.helpers import (log, get_exception_message, now)


class Server(object):
  def __init__(self, name, ip, username, password=None, key_path=None):
    self.port = 22
    if ":" in ip:
      self.port = int(ip.split(":")[-1])
      ip = ip.split(":")[0]

    import paramiko
    self.name = name
    self.ip = ip
    self.username = username
    self.password = password
    self.connect_tries = 0
    self.connected = False

    self.key = paramiko.RSAKey.from_private_key_file(
      key_path) if key_path else None

    self.ssh_client = paramiko.SSHClient()
    self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    self.sftp = None

  def connect(self):
    self.connect_tries += 1
    from scp import SCPClient

    try:
      self.ssh_client.connect(
        self.ip,
        self.port,
        self.username,
        self.password,
        timeout=4,
        pkey=self.key)
      self.sftp = self.ssh_client.open_sftp()
      self.scp = SCPClient(self.ssh_client.get_transport())
      self.connected = True
      log("Successful Connection to {} ({}).".format(self.name, self.ip))

    except Exception as E:
      log("Failed to connect to {} ({})".format(self.name, self.ip))
      log(E)
    finally:
      return self.connected

  def ssh_command(self, command, wait_for_output=True):
    if not self.connected: self.ssh_connect()

    try:
      stdin, stdout, stderr = self.ssh_client.exec_command(command)
    except socket.error as e:
      # not connected?
      self.connected = False
      self.ssh_connect()
      stdin, stdout, stderr = self.ssh_client.exec_command(command)

    self.last_output = ''
    self.last_output_lines = 0

    if (wait_for_output):
      self.ssh_chan_status = stdout.channel.recv_exit_status()
      for line in stdout.readlines():
        self.last_output = self.last_output + line
        self.last_output_lines += 1

    return self.last_output

  def sftp_download(self, remote_filepath, local_filepath, use_scp=True):
    if (not self.test_connection): self.ssh_connect()
    # copy file from remote server object to local path
    log("Downloading from '" + remote_filepath + "' to '" + local_filepath +
        "'")
    try:
      self.last_stat = None
      if use_scp:
        self.scp.get(remote_filepath, local_filepath)
      else:
        self.sftp.get(
          remote_filepath, local_filepath, callback=self.transfer_progress)
    except Exception as E:
      log('remote_filepath: {}:{}'.format(self.name, remote_filepath))
      log('local_filepath: {}'.format(local_filepath))
      log(E)

  def sftp_upload(self, local_filepath, remote_filepath, use_scp=True):
    if (not self.test_connection): self.ssh_connect()
    # copy file from local path to remote server object
    log("Uploading from '" + local_filepath + "' to '" + remote_filepath + "'")
    try:
      self.last_stat = None
      if use_scp:
        self.scp.put(local_filepath, remote_filepath)
      else:
        self.sftp.put(
          local_filepath, remote_filepath, callback=self.transfer_progress)
    except Exception as E:
      log('remote_filepath: {}:{}'.format(self.name, remote_filepath))
      log('local_filepath: {}'.format(local_filepath))
      log(E)

  def transfer_progress(self, transferred, total, unit='B'):
    "Display transfer progress"
    prct = int(100.0 * transferred / total)
    divide = lambda x, y: round(1.0 * x / (y), 1)

    if self.last_stat:
      secs = (datetime.datetime.now() - self.last_stat['time']).total_seconds()
      if secs > 2:
        rate = round((transferred - self.last_stat['transferred']) / secs, 1)
        self.last_stat = dict(time=now(), transferred=transferred, rate=rate)
      else:
        rate = self.last_stat['rate']
    else:
      rate = 0
      self.last_stat = dict(time=now(), transferred=transferred, rate=rate)

    if total > 1024**3:
      transferred = divide(transferred, 1024**3)
      total = divide(total, 1024**3)
      unit = 'GB'
      rate = '{} {} / sec'.format(divide(rate, 1024**2), 'MB')
    elif total > 1024**2:
      transferred = divide(transferred, 1024**2)
      total = divide(total, 1024**2)
      unit = 'MB'
      rate = '{} {} / sec'.format(divide(rate, 1024**2), unit)
    elif total > 1024**1:
      transferred = divide(transferred, 1024**1)
      total = divide(total, 1024**1)
      unit = 'KB'
      rate = '{} {} / sec'.format(divide(rate, 1024**1), unit)
    log('+{}% Complete: {} / {} {} @ {}'.format(
      prct, transferred, total, unit, rate))

  def test_connection(self):
    stdout = self.ssh_client.exec_command('ls')
    return self.connected



# ETL Library
import sys, os, requests, time, json, datetime, getpass, argparse

import copy, shutil

from multiprocessing import Queue, Process

from xutil.helpers import (
  load_profile,
  log, log,
  now,
  now_str,
  split,
  get_exception_message,
  get_databases,
  get_kw,
  get_db_profile,
  load_profile,
  struct)
from xutil.diskio import (
  read_yaml, read_file, write_csv,
  write_csvs, write_jsonl, write_jsonls,
  write_pqs
)

from xutil.web import send_email_exchange
import xutil.parallelism as parallelism

from xutil.database.base import (
  get_conn,
  DBConn)


from xutil.database.spark import (
  Spark,
  SparkConn,
)

from pathlib import Path



conns = {}


def exec_sql(
        database,
        sql=None,
        file_path=None,
        template_values=None,
        username=None,
        password=None,
        limit=None,
        delim=',',
        output_name=None,
        email=None,
        output_dir=None,
        query_name=None,
        echo=True,
        **kwargs
  ):
  """
  Function to execute SQL code on specified database

  Args:
      database: The database name to query (i.e: ORACLE_EDW, SPARK).
      sql: SQL snippet to execute.
      file_path: The source SQL file. Example: /path/to/file.sql
      template_values: A key / value list to fill a SQL template, separated by "&" and "=". i.e.:
      owner=USER1&tables=TABLE1,TABLE2
      username: The username to use to access database. Password will be prompted if not provided.
      password: password to use to access database.
      limit: he max number of records to pull.
      delim: The delimiter to use for output flat files. Default is ",".
      output_name: The file name to use. i.e.: "filename.csv".
      email: The email to send the output to. i.e.: "username@domain.com".
      output_dir: The output directory for the flat files. Default is "/path/to/folder".

  Returns:
      (field, rows) as a tuple if output_dir or email is not specified.
      field: is a list of the column names
      rows: is a list of tuples

  """
  global conns

  if file_path:
    sql = read_file(file_path)

  if template_values:
    template_values = template_values.split('&')
    kv_data = {kv.split('=')[0]: kv.split('=')[1] for kv in template_values}
    sql = sql.format(**kv_data)

  if output_dir:
    os.system("mkdir -p " + output_dir)

  if 'workflow' in kwargs and echo:
    print('***** EXECUTING SQL *****\n' + sql)

  dbs = get_databases()

  if database in get_databases():
    conn = get_conn(
      database,
      use_spark=True,
      echo=True,
    )
  else:
    log("Database is not supported!")
    log("Exiting.")
    sys.exit(1)

  conns[database] = conn
  fields, rows = conn.execute(
    sql,
    dtype='tuple',
    limit=limit,
    query_name=query_name,
    echo=echo
  )

  if fields and rows:

    if not (output_dir or email):
      return (fields, rows)

    user = os.getenv('USER')
    file_name = output_name if output_name else '{}-{}-{}.csv'.format(user, 'sql', int(time.time()))
    file_path = '{}/{}'.format(output_dir, file_name)
    write_csv(file_path, fields, rows, delim=delim)

    if email:
      if os.path.getsize(file_path) > 2 * (1024 ** 2):
        rc = os.system('gzip -f ' + file_path)
        file_path = file_path + '.gz' if rc == 0 else file_path

      subj = file_name
      body_text = 'See CSV attached\n\nROWS: {rows}\n\nSQL:\n{sql}'.format(rows=len(rows), sql=sql)

      send_email_exchange(
        to_address=email,
        subject="{} | {}".format(database, file_name),
        body_text=body_text,
        attachments=[file_path],
      )


def db_to_db(src_db,
             src_table,
             tgt_db,
             tgt_table,
             src_sql=None,
             where_clause='',
             tgt_mode='append',
             log=log,
             **kwargs):
  "Database to Database"
  print('= '*80)
  log('Launching data transfer Database to Database.')
  log('Source: {} | {}'.format(src_db, src_table))
  log('Target: {} | {}'.format(tgt_db, tgt_table))

  dbs = get_databases()
  if src_db not in dbs:
    raise Exception("Cannot find '{}' in database profile!".format(src_db))
  if tgt_db not in dbs:
    raise Exception("Cannot find '{}' in database profile!".format(tgt_db))

  src_db_prof = get_db_profile(src_db)
  tgt_db_prof = get_db_profile(tgt_db)

  use_jdbc = get_kw('use_jdbc', False, kwargs)
  hdfs_folder = get_kw('hdfs_folder', None, kwargs)
  partition_col = get_kw('partition_col', None, kwargs)
  partitions = int(get_kw('partitions', 5, kwargs))
  ff_only = get_kw('ff_only', False, kwargs)
  order_by = get_kw('order_by', [], kwargs)
  delete_after = get_kw('delete_after', False, kwargs)

  partition_col = partition_col if partition_col else get_partition_col(src_db, src_table)

  if tgt_db_prof.type in ('hive', 'spark') and src_db_prof.type not in (
      'hive', 'spark') and not use_jdbc:

    conn = get_conn(src_db)
    out_folder = conn.variables['tmp_folder']

    if src_table and partition_col and not src_sql:
      if ff_only:
        file_path = '{}/{}'.format(out_folder, src_table.lower())
      else:
        file_path, tot_cnt = db_table_to_ff_stream(
          src_db=src_db,
          table=src_table,
          partition_col=partition_col,
          where_clause=where_clause,
          out_folder=out_folder,
          partitions=partitions,
          max_running=partitions,
          log=log,
        )
    else:
      file_path = out_folder + '/' + tgt_table.lower() + '.csv'
      if not ff_only:
        src_sql = src_sql if src_sql else 'select * from ' + src_table
        data_strm = conn.stream(src_sql, rec_name=tgt_table)
        tot_cnt = write_csvs(file_path, data_strm, gzip=True, log=log)
      file_path = file_path + '.gz'

    ff_to_db(
      file_path,
      src_deli=',',
      src_timestamp_fmt='yyyy-MM-dd HH:mm:ss',
      src_date_fmt='yyyy-MM-dd',
      src_date_cols=[],
      src_pre_proc=None,
      tgt_db=tgt_db_prof.name,
      tgt_table=tgt_table,
      tgt_mode=tgt_mode,
      banner=False,
      log=log,
      delete_after=delete_after,
      hive_enabled=True,
      hdfs_folder=hdfs_folder,
      tot_cnt=tot_cnt,
    )
  else:
    # Use JDBC
    sparko = Spark(hive_enabled=False, spark_home=os.environ['SPARK_HOME'])
    if src_db_prof.type in ('hive', 'spark'):
      log("Reading Hive Table : " + src_table)
      sparko.sql('REFRESH TABLE ' + src_table)
      df1 = sparko.sql('select * from ' + src_table)
    else:
      src_table = '({})'.format(src_sql) if src_sql else src_table

      if partition_col:
        partition_dict = dict(
          partitionColumn=partition_col,
          numPartitions=partitions,
        )
      else:
        partition_dict = {}

      df1 = sparko.jdbc_read(
        src_db, src_table, partition_dict, log=log)
      tot_cnt = sparko._last_df_cnt

    sparko.jdbc_write(
      df1,
      tgt_db_prof,
      tgt_table,
      tgt_mode=tgt_mode,
      order_by=order_by,
      partitions=partitions,
      truncate=get_kw('truncate', None, kwargs),
      count_recs=get_kw('count_recs', None, kwargs),
      grant_sql=get_kw('grant_sql', None, kwargs),
      post_sql=get_kw('post_sql', None, kwargs),
      tot_cnt=tot_cnt,
      log=log)

  return dict(completed=True)



def db_to_ff(src_db, src_table, src_sql, tgt_ff, tgt_deli,
                 tgt_timestamp_fmt=None, tgt_date_fmt=None, where_clause='', log=log, **kwargs):
  "Database to Flat-File"
  if get_kw('banner', True, kwargs):
    print('= ' * 80)
    log('Launching data transfer Database to Flat-File.')
    log('Source: {} | {}'.format(src_db, src_table))
    log('Target: {}'.format(tgt_ff))

  dbs = get_databases()
  if src_db not in dbs:
    raise Exception("Cannot find '{}' in database profile!".format(src_db))

  src_db_prof = get_db_profile(src_db)

  partition_col = get_kw('partition_col', None, kwargs)
  stream_partitions = int(get_kw('stream_partitions', 5, kwargs))
  partitions = int(get_kw('partitions', 1, kwargs)) # final number of partitions


  if src_db_prof.type not in ('spark', 'hive'):
    sparko = Spark(hive_enabled=False)
    src_table = '({})'.format(src_sql) if src_sql else src_table
    if partition_col:
      partition_dict = dict(
        partitionColumn=partition_col, numPartitions=stream_partitions)
    else:
      partition_dict = {}
    df1 = sparko.jdbc_read(
      src_db, src_table, partition_dict, log=log)
    tot_cnt = sparko._last_df_cnt
  else:
    sparko = Spark(hive_enabled=True)
    log("Reading Hive Table : " + (src_table if src_table else 'QUERY'))
    sql = src_sql or 'select * from ' + src_table
    df1 = sparko.sql(sql)

  sparko.write_csv2(
    df1,
    tgt_ff,
    timestampFormat=tgt_timestamp_fmt,
    dateFormat=tgt_date_fmt,
    delimeter=tgt_deli,
    gzip=kwargs['gzip'] if 'gzip' in kwargs else True,
    partitions=partitions,
    log=log,
  )

  if 'tgt_post_proc' in kwargs and kwargs['tgt_post_proc'] in globals():
    globals()[kwargs['tgt_post_proc']](tgt_ff, log=log)

  return dict(completed=True)




def ff_to_db(src_ff, src_deli, tgt_db, tgt_table,
                 src_timestamp_fmt=None, src_date_fmt=None,
                 tgt_mode='append', log=log, **kwargs):
  "Flat-File to Database"
  if get_kw('banner', True, kwargs):
    print('= ' * 80)
    log('Launching data transfer Flat-File to Database.')
    log('Source: {}'.format(src_ff))
    log('Target: {} | {}'.format(tgt_db, tgt_table))

  if tgt_db not in get_databases():
    raise Exception("Cannot find '{}' in database profile!".format(tgt_db))

  tgt_db_prof = get_db_profile(tgt_db)

  hive_enabled = get_kw('hive_enabled', None, kwargs)
  hdfs_folder = get_kw('hdfs_folder', None, kwargs)
  src_date_cols = get_kw('src_date_cols', None, kwargs)
  order_by = get_kw('order_by', [], kwargs)
  src_pre_proc = get_kw('src_pre_proc', None, kwargs)
  delete_after = get_kw('delete_after', False, kwargs)

  if src_pre_proc and src_pre_proc in globals():
    globals()[src_pre_proc](src_ff, log=log)

  if tgt_db_prof.type in ('spark', 'hive'):
    conn = get_conn(tgt_db_prof.name)
    sparko = conn.sparko
  else:
    sparko = Spark(hive_enabled=False)
  log("Loading file {}..".format(src_ff))

  df1 = sparko.read_csv2(
    src_ff,
    delimeter=src_deli,
    timestampFormat=src_timestamp_fmt,
    dateFormat=src_date_fmt,
    date_cols=src_date_cols,
    hdfs_folder=hdfs_folder,
    log=log,
  )

  # To Hive
  if tgt_db_prof.type in ('hive', 'spark'):
    sparko.hive_write(df1, tgt_table, order_by=order_by, mode=tgt_mode, log=log)
    if delete_after:
      os.system('rm -rf ' + src_ff)

  # To RDBMS, use JDBC
  else:
    sparko.jdbc_write(
      df1,
      tgt_db_prof,
      tgt_table,
      tgt_mode=tgt_mode,
      order_by=order_by,
      partitions=get_kw('partitions', 5, kwargs),
      truncate=get_kw('truncate', None, kwargs),
      count_recs=get_kw('count_recs', None, kwargs),
      grant_sql=get_kw('grant_sql', None, kwargs),
      post_sql=get_kw('post_sql', None, kwargs),
      tot_cnt=get_kw('tot_cnt', 0, kwargs),
      log=log)

  return dict(completed=True)


def ff_to_ff(src_ff, src_deli,
                 src_timestamp_fmt, src_date_fmt, src_date_cols,
                 tgt_ff, tgt_deli, tgt_timestamp_fmt, tgt_date_fmt, log=log, **kwargs):
  "Flat-File to Flat-File"

  if 'src_pre_proc' in kwargs and kwargs['src_pre_proc'] in globals():
    globals()[kwargs['src_pre_proc']](src_ff, log=log)

  sparko = Spark(hive_enabled=False)

  # read
  df1 = sparko.read_csv2(
    src_ff,
    delimeter=src_deli,
    timestampFormat=src_timestamp_fmt,
    dateFormat=src_date_fmt,
    date_cols=src_date_cols,
    log=log,
  )

  # Write
  sparko.write_csv2(
    df1,
    tgt_ff,
    timestampFormat=tgt_timestamp_fmt,
    dateFormat=tgt_date_fmt,
    delimeter=tgt_deli,
    log=log,
  )

  if 'tgt_post_proc' in kwargs and kwargs['tgt_post_proc'] in globals():
    globals()[kwargs['tgt_post_proc']](tgt_ff, log=log)

  return dict(completed=True)


def bash(cmd, *args, **kwargs):
  process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
  for line in iter(process.stdout.readline, b''):
     sys.stdout.write(line.decode())
     sys.stdout.flush()


def get_partition_col(src_db, table, n=20000, echo=False):
  """Automatically detect a suitable partition column with n sample records"""
  conn = get_conn(src_db)
  field_rows = conn.get_columns(table, native_type=False)
  fields = [r.column_name.lower() for r in field_rows]
  number_fields_i = [i for i,r in enumerate(field_rows) if r.type in ('integer', 'double', 'decimal')]
  date_fields_i = [i for i,r in enumerate(field_rows) if r.type in ('datetime')]
  fields_i = date_fields_i + number_fields_i
  fields_stat = {
    fields[fi]: dict(nulls=0, uniques=set([]), min=None, max=None, ii=ii)
    for ii, fi in enumerate(fields_i)
  }

  date_conv = lambda f: conn.template('function.date_to_int').format(field=f)
  num_conv = lambda f: conn.template('function.number_to_int').format(field=f)
  fields_expr = [date_conv(fields[i])
                for i in date_fields_i] + [num_conv(fields[i]) for i in number_fields_i]
  fields_sql = ', '.join('{} as {}'.format(fexpr, fields[fields_i[ii]]) for ii, fexpr in enumerate(fields_expr))
  try:
    sql = conn.template('core.sample').format(fields=fields_sql, table=table, n=n)
    data = conn.select(sql, dtype='tuple', echo=echo)
  except Exception as E:
    err_msg = get_exception_message(raw=True).lower()
    if 'cannot select' in err_msg:
      log('-'+err_msg)
      log('-Retrying without Sampling')
      sql = conn.template('core.limit').format(
        fields=fields_sql, table=table, n=n)
      data = conn.select(sql, dtype='tuple', echo=echo)
    else:
      raise E


  fields = [f.lower() for f in conn._fields]

  if len(data) < n-2:
    return None

  for row in data:
    for fi, val in enumerate(row):
      f = fields[fi]
      fields_stat[f]['uniques'].add(val)
      if val is None:
        fields_stat[f]['nulls'] += 1
      else:
        if fields_stat[f]['min'] is None or val < fields_stat[f]['min']:
          fields_stat[f]['min'] = val
        if fields_stat[f]['max'] is None or val > fields_stat[f]['max']:
          fields_stat[f]['max'] = val

  for f in fields:
    fields_stat[f]['gap'] = float(fields_stat[f]['max'] - fields_stat[f]['min']) if fields_stat[f]['max'] and fields_stat[f]['min'] else 0
    fields_stat[f]['uniques_cnt'] = len(fields_stat[f]['uniques'])
    fields_stat[f][
      'uniques_prct'] = 100.0 * fields_stat[f]['uniques_cnt'] / len(data)
    fields_stat[f]['nulls_prct'] = 100.0 * fields_stat[f]['nulls'] / len(data)
    fields_stat[f]['count'] = len(data) - fields_stat[f]['nulls']
    # highest unique, lowest nulls
    fields_stat[f]['score'] = (100 - fields_stat[f]['nulls_prct']
                               ) * fields_stat[f]['uniques_prct'] / 100.0
    fields_stat[f][
      'score'] = 0 if fields_stat[f]['gap'] > 1000000 else fields_stat[f][
        'score']
    if echo:
      print('"{}" -> U{} | G{} | S{}'.format(f, fields_stat[f]['uniques_cnt'],
                                             fields_stat[f]['gap'],
                                             fields_stat[f]['score']))

  best_field = sorted(fields_stat, reverse=True, key=lambda f: fields_stat[f]['score'])[0]
  best_field_expr = fields_expr[fields_stat[best_field]['ii']]

  log('+Using partition col: '+best_field_expr)
  return best_field_expr



def get_sql_table_split(src_db, table, partition_col, partitions=10,
  where_clause='', cov_utf8=True, echo=False,):

  ######## Split and group the dates

  conn = get_conn(src_db, echo=echo, reconnect=True)

  split_data = conn.select(
    conn.template('routine.number_trunc_uniques').format(
      partition_col_trunc=conn.template('function.truncate_f').format(
        field=partition_col),
      where='where ' + where_clause if where_clause else '',
      table=table),
    rec_name=table if ' ' in partition_col else partition_col,
    echo=False)

  groups = []
  curr_group = []
  curr_group_size = 0
  total_cnt = sum([row.cnt for row in split_data])

  part_size = int(total_cnt / partitions)

  log('total_cnt is {}'.format(total_cnt))
  log('part_size is {} X {}'.format(part_size, partitions))

  ######## number of groups should equal partitions
  for row in split_data:
    val = row.trunc_field.strftime('%Y-%m-%d') if isinstance(
      row.trunc_field, datetime.datetime) else row.trunc_field
    curr_group.append(row.trunc_field)
    curr_group_size += row.cnt
    if curr_group_size > part_size:
      if curr_group_size > part_size * 2:
        log('-WARNING: Skewed data partition. curr_group_size({}) >  part_size * 2'.format(curr_group_size))
      groups.append(curr_group)
      curr_group = []
      curr_group_size = 0
  if curr_group:
    groups.append(curr_group)

  ######## get fields
  fields_rows = conn.get_columns(table, echo=False)
  fields_conv = []
  fields = []

  for row in fields_rows:
    field = row.column_name.lower()
    fields.append(field)

    if conn.type in ('oracle') and ('DATE' in row.type or 'TIME' in row.type
                                    or 'LONG' in row.type):
      fields_conv.append(field)
    elif conn.type in ('oracle') and cov_utf8:
      # to mitigate ORA-29275: partial multibyte character
      fields_conv.append((conn.template('function.str_utf8') +
                          " as {field}").format(field=field))
    else:
      fields_conv.append(row.column_name.lower())

  sql_templ = conn.template('routine.number_trunc_min_max')
  sqls = []
  for i, group in enumerate(groups):
    sql = sql_templ.format(
      table=table,
      fields=', '.join(fields_conv),
      where=where_clause + ' and ' if where_clause else '',
      partition_col_trunc=conn.template('function.truncate_f').format(
        field=partition_col),
      min_val=min([val for val in group if val]),
      max_val=max([val for val in group if val]),
      or_null=' or {} is null'.format(partition_col) if None in group else '')
    sqls.append(sql)

  return sqls, fields, total_cnt


def db_table_to_ff_stream(src_db,
                     table,
                     partition_col,
                     out_folder,
                     out_type='parquet',
                     where_clause='',
                     partitions=10,
                     file_name=None,
                     log=log,
                     echo=True,
                     **kwargs):
  s_t = datetime.datetime.now()
  log('Reading table "{}"'.format(table))

  gzip = get_kw('gzip', True, kwargs)
  cov_utf8 = get_kw('cov_utf8', True, kwargs)
  max_running = get_kw('max_running', partitions, kwargs)


  def sql_to_csv(src_db, sql, file_path, fields, queue=None, gzip=True):
    def log2(text, color=None):
      queue.put(('log', text))

    log2 = log2 if queue else log

    conn = get_conn(src_db, echo=False, reconnect=True)

    try:
      counter = write_csvs(
        file_path,
        conn.stream(sql, dtype='tuple', echo=False),
        fields=fields,
        gzip=gzip,
        log=log2)

      queue.put(('wrote-counter', counter))

    except Exception as E:
      log(E)
      queue.put(('exception', E))

  def sql_to_pq(src_db, sql, file_path, queue=None):
    def log2(text, color=None):
      queue.put(('log', text))

    log2 = log2 if queue else log

    conn = get_conn(src_db, echo=False, reconnect=True)

    try:
      counter = write_pqs(
        file_path,
        conn.stream(sql, dtype='dataframe', yield_chuncks=True, chunk_size=200000, echo=False),
        append=True,
        log=log2)

      queue.put(('wrote-counter', counter))

    except Exception as E:
      log(E)
      queue.put(('exception', E))

  # Split SELECT sql statements
  sqls, fields, total_cnt = get_sql_table_split(
    src_db,
    table,
    partition_col,
    partitions=partitions,
    where_clause=where_clause,
    cov_utf8=cov_utf8,
    echo=echo)

  ######## delete if exists and create folder
  table_folder = '{}/{}'.format(out_folder, file_name
                                if file_name else table.lower())
  os.system('rm -rf {f}; mkdir {f}'.format(f=table_folder))
  
  if os.path.exists(table_folder):
    shutil.rmtree(table_folder, ignore_errors=True)
  os.mkdir(table_folder)


  ######## Start the processes
  procs = []
  remaining_procs = []
  file_parts = []

  proc_items = {}
  for i, sql in enumerate(sqls):
    file_path = '{}/{}.{}'.format(table_folder, table.lower(),
                                     format(i, "03d"))
    file_parts.append(file_path)
    queue = Queue()
    if out_type == 'parquet':
      _proc = Process(
      target=sql_to_pq, args=(src_db, sql, table_folder, queue))
    else:
      _proc = Process(
      target=sql_to_csv, args=(src_db, sql, file_path + '.csv', fields, queue, gzip))

    proc_items[file_path] = dict(
      sql=sql, proc=_proc, queue=queue, prog_cnt=0)

    if len(procs) == max_running:
      remaining_procs.append(_proc)
    else:
      procs.append(_proc)
      procs[-1].start()

  ######## Keep track of streams
  done = []
  get_running_procs = lambda: [proc for proc in procs if proc.is_alive()]
  last_dt = datetime.datetime.now() - datetime.timedelta(seconds=30)
  last_prog_cnt = 0
  wrote_cnt = 0
  rate = 10000
  exptn = None

  while len(done) < len(file_parts):
    time.sleep(0.1)
    secs = round((datetime.datetime.now() - s_t).total_seconds(), 1)
    running_procs = get_running_procs()

    for file_path in proc_items:
      if not proc_items[file_path]['queue'].empty():
        # key can be log, counter, exception
        key, val = proc_items[file_path]['queue'].get()
        if key == 'wrote-counter':
          wrote_cnt += val
          done.append(file_path)
        if key == 'exception':
          log('Process exception for "{}":\n{}'.format(file_path, val))
          raise (val)
        if key == 'log':
          if 'rows' in val:
            proc_items[file_path]['prog_cnt'] = int(
              val.split(':')[1].split('rows')[0])

    secsd = (datetime.datetime.now() - last_dt).total_seconds()
    if secsd > 15:  # update each 15 sec
      prog_cnt = sum([item['prog_cnt'] for item in proc_items.values()])
      last_dt = datetime.datetime.now()
      rate = round((prog_cnt - last_prog_cnt) / secsd, 1)
      eta_mins = round((total_cnt - prog_cnt) / rate / 60,
                       1) if rate > 0 else '~'
      last_prog_cnt = prog_cnt

      text_log = '{elapsed_mins} m -- Streaming: {prog_cnt} / {total_cnt} [{rate} r/s] [{prct} %] --  {done_cnt} / {file_cnt} finished. {running_cnt} running. ETA is {eta_mins} mins'.format(
        elapsed_mins=round(secs / 60, 1),
        prog_cnt=prog_cnt,
        total_cnt=total_cnt,
        rate=rate,
        prct=round(100.0 * prog_cnt / total_cnt, 1),
        done_cnt=len(done),
        file_cnt=len(file_parts),
        running_cnt=len(running_procs),
        eta_mins=eta_mins)
      log(text_log)

    if len(running_procs) < max_running and remaining_procs:
      _proc = remaining_procs.pop(0)
      _proc.start()
      procs.append(_proc)

  secs = (datetime.datetime.now() - s_t).total_seconds()
  rate = round(wrote_cnt / secs, 1)
  if echo:
    log(" >>> Wrote {} rows in {} secs [{} r/s] to {}".format(
      wrote_cnt, secs, rate, table_folder))

  if wrote_cnt != total_cnt:
    log(
      '-WARNING: written count ({}) is different from initial total cnt ({}).'.
      format(wrote_cnt, total_cnt))
    log('-WARNING: total_cnt - wrote_cnt = {}'.format(total_cnt - wrote_cnt))

  return table_folder




etl_usage = '''
exec_etl.py <command> [<args>]
 
ETL commands available:
db-db     Database to Database transfer
ff-db     Flat-File to Database transfer
db-ff     Database to Flat-File transfer
 
Examples:
python exec_etl.py db-db --src_db=DB1 --src_table=schema_name1.table_name --tgt_db=HIVE --tgt_table=schema_name2.table_name
python exec_etl.py db-ff --src_db=DB1 --src_table=schema_name1.table_name --tgt_ff=/path/of/file.csv
python exec_etl.py ff-db --src_ff=/path/of/file.csv --tgt_db=HIVE --tgt_table=schema_name2.table_name
'''



class SqlCmdParser(object):
  def __init__(self):

    parser = argparse.ArgumentParser(description='Execute SQL from file or command line argument. Optional Output to Flat file.')
    parser.add_argument('--sql', help='SQL snippet to execute. Use double quote(").')
    parser.add_argument('--file_path', help='The source SQL file. Example: /path/to/file.sql')
    parser.add_argument('--template_values', help='A key / value list to fill a SQL template, separated by "&" and "=". i.e.: owner=SCHEMA_NM&tables=INVOICES,VENDORS')
    parser.add_argument('--database', help='The database name to query (i.e: EDW_1, SPARK).')
    parser.add_argument('--user', help='The username to use to access database. Password will be prompted.')
    parser.add_argument('--limit', help='The max number of records to pull.')
    parser.add_argument('--delim', help='The delimiter to use for output flat files. Default is ",".')
    parser.add_argument(
      '--output_dir',
      help='The output directory for the flat files. Default is "/tmp".')
    parser.add_argument('--output_name', help='The file name to use. i.e.: "Invoices_flat_wide.csv".')
    parser.add_argument('--email', help='The email to send the output to. i.e.: "tom@mydomain.com".')

    args = parser.parse_args()

    if not (args.database and (args.sql or args.file_path)):
      log("Need to specify database and (sql or filepath).", color='red')
      log("""Example: python /data/public/ds/scripts/exec-sql.py --sql='show databases;' --database=SPARK --output_name='Hive_databases.txt'""", color='green')
      parser.print_help()
      sys.exit(1)

    if args.database != 'SPARK' and not args.user:
      log("Need to specify user name with flag --user", color='red')
      log("Exiting.", color='red')
      sys.exit(1)

    exec_sql(
      database = args.database,
      file_path = args.file_path,
      sql = args.sql,
      template_values = args.template_values,
      username = args.user,
      limit = int(args.limit) if args.limit else None,
      delim = args.delim if args.delim else ',',
      output_name = args.output_name,
      email = args.email,
      output_dir = args.output_dir if args.output_dir else "/tmp",
    )


class EtlCmdParser(object):
  def __init__(self):
    parser = argparse.ArgumentParser(
      description='Execute ETL operations though CLI.', usage=etl_usage)
    parser.add_argument('command', help='ETL Subcommand to run')

    # parse_args defaults to [1:] for args, but you need to
    # exclude the rest of the args too, or validation will fail
    args = parser.parse_args(sys.argv[1:2])
    command = args.command.replace('-', '_')

    if not hasattr(self, command):
      print('Unrecognized command!')
      parser.print_help()
      exit(1)

    # use dispatch pattern to invoke method with same name
    getattr(self, command)()

  def db_db(self):
    "Database to Database"
    parser = argparse.ArgumentParser(
      description='Database to Database transfer')

    # prefixing the argument with -- means it's optional
    parser.add_argument('--src_db', help='The source database name.')
    parser.add_argument(
      '--src_table', help='The source table name (including the schema).')
    parser.add_argument(
      '--src_sql',
      help=
      'The SQL snippet to execute from the source database. Use double quote(").'
    )
    parser.add_argument(
      '--src_sql_file',
      help=
      'The source SQL file to execute from the source database. Example: /path/to/file.sql'
    )
    parser.add_argument('--tgt_db', help='The target database name.')
    parser.add_argument(
      '--tgt_table', help='The target table name (including the schema).')
    parser.add_argument(
      '--append',
      help=
      'Append to the target table. Default will drop and overwrite the target table.'
    )

    # now that we're inside a subcommand, ignore the first
    args = parser.parse_args(sys.argv[2:])

    if not (args.src_db and
            (args.src_sql or args.src_sql_file or args.src_table)
            and args.tgt_db and args.tgt_table):
      log(
        "Need to specify src_db and (sql or filepath or src_table) and tgt_db and tgt_table."
      )
      parser.print_help()
      sys.exit(1)

    if args.src_sql_file:
      src_sql = read_file(args.src_sql_file)

    db_to_db(
      src_db=args.src_db,
      src_table=args.src_table,
      src_sql=src_sql,
      tgt_db=args.tgt_db,
      tgt_table=args.tgt_table,
      tgt_mode='append' if args.tgt_mode else 'overwrite',
    )

  def ff_db(self):
    "Flat File to Database"
    parser = argparse.ArgumentParser(
      description='Flat File to Database transfer')

    # prefixing the argument with -- means it's optional
    parser.add_argument('--src_ff', help='The source file path.')
    parser.add_argument(
      '--delim',
      help='The delimiter to use to parse the file. Default is ",".')
    parser.add_argument(
      '--datetime_format',
      help=
      'Datetime Format throughout the file. Default is "yyyy-MM-dd HH:mm:ss".')
    parser.add_argument(
      '--datetime_fields',
      help=
      'Datetime fields that should be parsed with specified datetime-format. ie: "date_field_1,date_field_2"'
    )
    parser.add_argument('--tgt_db', help='The target database name.')
    parser.add_argument(
      '--tgt_table', help='The target table name (including the schema).')
    parser.add_argument(
      '--tgt_mode', help='The target table insert mode (overwrite or append).')
    parser.add_argument(
      '--hdfs_folder',
      help=
      'The temporary HDFS field that the field should be stored in. Default is "/tmp"'
    )
    parser.add_argument(
      '--append',
      help='Append to existing table. Default is "overwrite".',
      action='store_true')
    
    parser.add_argument(
      '--remove_cr',
      help='Remove carriage return from Flat file.',
      action='store_true')

    # now that we're inside a subcommand, ignore the first
    args = parser.parse_args(sys.argv[2:])

    if not (args.src_ff and args.tgt_db and args.tgt_table):
      log("Need to specify src_ff and tgt_db and tgt_table.")
      parser.print_help()
      sys.exit(1)
    
    if args.remove_cr:
      remove_cr(args.src_ff)

    ff_to_db(
      src_ff=args.src_ff,
      src_deli=args.delim if args.delim else ',',
      src_timestamp_fmt=args.datetime_format if args.datetime_format else None,
      src_date_cols=args.datetime_fields.split(',')
      if args.datetime_fields else [],
      tgt_db=args.tgt_db,
      tgt_table=args.tgt_table,
      tgt_mode='append' if args.tgt_mode else 'overwrite',
      hdfs_folder=args.hdfs_folder if args.hdfs_folder else '/tmp',
    )

  def db_ff(self):
    "Database to Flat File"
    parser = argparse.ArgumentParser(description='Database to Flat File')

    # prefixing the argument with -- means it's optional
    parser.add_argument('--src_db', help='The source database name.')
    parser.add_argument(
      '--src_table', help='The source table name (including the schema).')
    parser.add_argument(
      '--src_sql',
      help=
      'The SQL snippet to execute from the source database. Use double quote(").'
    )
    parser.add_argument(
      '--src_sql_file',
      help=
      'The source SQL file to execute from the source database. Example: /path/to/file.sql'
    )
    parser.add_argument('--tgt_ff', help='The target file path.')
    parser.add_argument(
      '--delim',
      help='The delimiter to use to parse the file. Default is ",".')
    parser.add_argument(
      '--datetime_format',
      help=
      'Datetime Format throughout the file. Default is "yyyy-MM-dd HH:mm:ss".')

    # now that we're inside a subcommand, ignore the first
    args = parser.parse_args(sys.argv[2:])

    if not (args.src_db and
            (args.src_sql or args.src_sql_file or args.src_table)
            and args.tgt_ff):
      log(
        "Need to specify src_db and (sql or filepath or src_table) and tgt_ff."
      )
      parser.print_help()
      sys.exit(1)

    if args.src_sql_file:
      src_sql = read_file(args.src_sql_file)

    db_to_ff(
      src_db=args.src_db,
      src_table=args.src_table if args.src_table else None,
      src_sql=args.src_sql if args.src_sql else None,
      tgt_ff=args.tgt_ff,
      tgt_deli=args.delim if args.delim else ',',
      tgt_date_fmt=args.datetime_format if args.datetime_format else None,
      tgt_timestamp_fmt=args.datetime_format if args.datetime_format else None,
    )


def clean_file_1(file_path, log=log):
  log('Cleaning file {}'.format(file_path))
  os.system('sed -i "1s/ /_/g" {}'.format(file_path))  # fix first row
  os.system('sed -i "s/\\tNone/\\t/g" {}'.format(file_path))  # fix whole file


def remove_cr(file_path, log=log):
  log('Removing \\r character from {}'.format(file_path))
  os.system('''perl -i -pe 's/\\r//' {}'''.format(file_path))  # fix whole file


def gzip_file(file_path, log=log):
  log('Gzipping {}'.format(file_path))
  os.system('gzip -f {}'.format(file_path))


class Project:
  """
  Reprensents an ETL project, intaking a YAML file.
  """

  def __init__(self, name, path, **kwargs):
    self.name = name
    self.yaml_path = path
    self.data = read_yaml(path)
    self.variables = self.data['variables']
    self.echo = kwargs['echo'] if 'echo' in kwargs else False

    self.project_name = self.data['project'].replace(' ', '_')
    self.parse_tasks(echo=self.echo)

  def parse_tasks(self, echo=False):
    tasks = {}

    def prep_item(item, echo=False):
      # apply template as applicable
      if 'template' in item:
        for k in self.variables[item['template']]:
          item[k] = item.get(k, self.variables[item['template']][k])

      worker_name = 'worker-' + item['from'].lower() if 'from' in item else None

      # add missing keys
      item['naming_prefix'] = item.get('naming_prefix', '')
      item['naming_include_schema'] = item.get('naming_include_schema', True)
      item['tgt_schema'] = item.get('tgt_schema', '')
      item['date_column_map'] = item.get('date_column_map', '')
      item['order_column_map'] = item.get('order_column_map', '')
      item['where_clause'] = item.get('where_clause', '')
      item['where_clause_template'] = item.get('where_clause_template', '')
      item['grant_sql_template'] = item.get('grant_sql_template', '')
      item['post_sql_template'] = item.get('post_sql_template', '')
      item['gzip'] = item.get('gzip', False)
      item['folder'] = item.get('folder', None)
      item['grant_sql'] = item.get('grant_sql', '')
      item['post_sql'] = item.get('post_sql', '')
      item['partitions'] = item.get('partitions', 1)
      item['worker_filter'] = item.get('worker_filter', worker_name)
      item['order_field'] = []
      item['partition_col'] = None
      item['mode'] = item.get('mode', 'overwrite')
      item['delete_after'] = item.get('delete_after', False)

      item['pre_processor'] = item.get('pre_processor', '')
      item['post_processor'] = item.get('post_processor', '')

      return item

    # db-to-db ###############
    def process_db_to_db(echo=False):
      if echo: log('process_db_to_db')
      typ = 'db_to_db'

      for item in self.data[typ]:

        if echo: log('processing item: ' + str(item))
        item_ = prep_item(item)

        if 'table_list' in item:
          for n, table in enumerate(item_['table_list']):
            where_clause = None
            table_dict = table if isinstance(table, dict) else None

            if table_dict:
              where_clause = table_dict['where_clause'] if 'where_clause' in table_dict else where_clause
              table = table_dict['table']
            
            item = copy.deepcopy(item_)
            shema_nm = table.split('.')[0].replace('[', '').replace(']', '')
            table_nm = table.split('.')[-1]
            name = item['workflow'] + '.' + table_nm.lower() if 'workflow' in item else typ + '.' + table.lower()

            if echo: log('processing task: ' + name)

            tgt_table = '{}_{}'.format(shema_nm, table_nm).lower() if item[
              'naming_include_schema'] else table_nm.lower()
            tgt_table = item['naming_prefix'] + tgt_table
            tgt_table = table_dict['tgt_table'] if table_dict and 'tgt_table' in table_dict else tgt_table
            item['tgt_table'] = item['tgt_table'] if 'tgt_table' in item else tgt_table
            item['tgt_table'] = '{}.{}'.format(item['tgt_schema'], item['tgt_table']) if item['tgt_schema'] else item[
              'tgt_table']

            if echo: log('tgt_table: ' + item['tgt_table'])
            item['query'] = item['query'] if 'query' in item else None

            if item['date_column_map'] and 'column_map_date' in self.variables:
              item['partition_col'] = self.variables['column_map_date'][table] if table in self.variables[
                'column_map_date'] else None

            if item['order_column_map'] and 'column_map_order' in self.variables:
              item['order_field'] = self.variables['column_map_order'][table] if table in self.variables[
                'column_map_order'] else None

            if item['date_field'] and item['where_clause_template'] and item['where_clause_template'] in self.variables:
              where_clause = self.variables[item['where_clause_template']].format(
                col=item['partition_col'],
                start_date=self.variables['start_date'],
                end_date=self.variables['end_date'],
              )

            if item['where_clause'] and not where_clause:
              where_clause = item['where_clause']

            item['grant_sql'] = item['grant_sql_template'].format(table=item['tgt_table']) if item[
              'grant_sql_template'] else item['grant_sql']
            item['post_sql'] = item['post_sql_template'].format(table=item['tgt_table']) if item[
              'post_sql_template'] else item['post_sql']

            if name in tasks:
              raise Exception('workflow "{}" already in tasks data!'.format(name))
            else:
              tasks[name] = dict(
                n=n + 1,
                workflow=item['workflow'],
                name=name,
                type=typ,
                src_table=table,
                src_db=item['from'],
                tgt_db=item['to'],
                src_sql=item['query'],
                tgt_table=item['tgt_table'],
                tgt_mode=item['mode'],
                partition_col=item['partition_col'],
                order_by=item['order_field'],
                where_clause=where_clause,
                grant_sql=item['grant_sql'],
                post_sql=item['post_sql'],
                partitions=item['partitions'],
                worker_filter=item['worker_filter'],
                delete_after=item['delete_after'],
                _item=item,
              )

        elif 'query' in item and 'tgt_table' in item:
          item = copy.deepcopy(item_)
          n = 0
          table = item['tgt_table']
          name = item['workflow'] + '.' + table.lower() if 'workflow' in item else typ + '.' + table.lower()
          if echo: log('processing task: ' + name)

          item['tgt_table'] = '{}.{}'.format(item['tgt_schema'], item['tgt_table']) if item['tgt_schema'] else item[
            'tgt_table']

          item['query'] = item['query'] if 'query' in item else None
          if item['date_column_map'] and 'column_map_date' in self.variables:
            item['partition_col'] = self.variables['column_map_date'][table] if table in self.variables[
              'column_map_date'] else None

          if name in tasks:
            raise Exception('workflow "{}" already in tasks data!'.format(name))
          else:
            tasks[name] = dict(
              n=n + 1,
              workflow=item['workflow'],
              name=name,
              type=typ,
              src_table=table,
              src_db=item['from'],
              tgt_db=item['to'],
              src_sql=item['query'],
              tgt_table=item['tgt_table'],
              partition_col=item['partition_col'],
              order_by=item['order_field'],
              where_clause=item['where_clause'],
              grant_sql=item['grant_sql'],
              partitions=item['partitions'],
              worker_filter=item['worker_filter'],
              tgt_mode=item['mode'],
              delete_after=item['delete_after'],
              _item=item,
            )

        else:
          raise Exception('Item did not meet any condition: ' + str(item))

    def process_ff_to_db(echo=False):
      typ = 'ff_to_db'
      for n, item in enumerate(self.data[typ]):
        name = item['workflow'] + '.' + item['tgt_table'].lower() if 'workflow' in item else typ + '.' + item[
          'tgt_table'].lower()
        if echo: log('processing task: ' + name)

        item = prep_item(item)
        tgt_table = '{}.{}'.format(item['tgt_schema'], item['tgt_table'])

        tasks[name] = dict(
          n=n + 1,
          workflow=item['workflow'],
          name=name,
          type=typ,
          src_ff=item['file_path'],
          src_deli=item['delimiter'],
          src_timestamp_fmt=item['timestamp_format'],
          src_date_fmt=item['date_format'],
          src_date_cols=item['date_cols'],
          src_pre_proc=item['pre_processor'],
          order_by=item['order_by'],
          tgt_db=item['to'],
          tgt_table=tgt_table,
          tgt_mode=item['mode'],
          _item=item,
        )

    def process_db_to_ff(echo=False):
      typ = 'db_to_ff'
      for item in self.data[typ]:
        if echo: log('processing item: ' + str(item))
        item_ = prep_item(item)

        if 'table_list' in item:
          for n, table in enumerate(item['table_list']):
            item = copy.deepcopy(item_)
            shema_nm, table_nm = table.split('.')
            name = item['workflow'] + '.' + table_nm.lower() if 'workflow' in item else typ + '.' + table.lower()
            if echo: log('processing task: ' + name)

            item['query'] = item['query'] if 'query' in item else None
            if item['date_column_map'] and 'column_map_date' in self.variables:
              item['partition_col'] = self.variables['column_map_date'][table] if table in self.variables[
                'column_map_date'] else None

            file_name = '{}_{}'.format(shema_nm, table_nm).lower() if item[
              'naming_include_schema'] else table_nm.lower()
            file_name = item['naming_prefix'] + file_name
            item['file_path'] = '{}/{}.csv'.format(item['folder'], file_name)

            tasks[name] = dict(
              n=n + 1,
              workflow=item['workflow'],
              name=name,
              type=typ,
              src_db=item['from'],
              src_table=table,
              src_sql=item['query'],
              tgt_ff=item['file_path'],
              tgt_deli=item['delimiter'],
              tgt_timestamp_fmt=item['timestamp_format'],
              tgt_date_fmt=item['date_format'],
              tgt_post_proc=item['post_processor'],
              partition_col=item['partition_col'],
              partitions=item['partitions'],
              gzip=item['gzip'],
              worker_filter=item['worker_filter'],
              _item=item,
            )
        elif 'query' in item:
          item = copy.deepcopy(item_)
          n = 0
          table = item['tgt_table']
          name = item['workflow'] + '.' + table.lower() if 'workflow' in item else typ + '.' + table.lower()
          if echo: log('processing task: ' + name)

          item['tgt_table'] = '{}.{}'.format(item['tgt_schema'], item['tgt_table']) if item['tgt_schema'] else item[
            'tgt_table']

          item['query'] = item['query'] if 'query' in item else None
          if item['date_column_map'] and 'column_map_date' in self.variables:
            item['partition_col'] = self.variables['column_map_date'][table] if table in self.variables[
              'column_map_date'] else None

          if name in tasks:
            raise Exception('workflow "{}" already in tasks data!'.format(name))
          else:
            tasks[name] = dict(
              n=n + 1,
              workflow=item['workflow'],
              name=name,
              type=typ,
              src_db=item['from'],
              src_table=table,
              src_sql=item['query'],
              tgt_ff=item['file_path'],
              tgt_deli=item['delimiter'],
              tgt_timestamp_fmt=item['timestamp_format'],
              tgt_date_fmt=item['date_format'],
              tgt_post_proc=item['post_processor'],
              partition_col=item['partition_col'],
              partitions=item['partitions'],
              worker_filter=item['worker_filter'],
              _item=item,
            )
        else:
          raise Exception('Item did not meet any condition: ' + str(item))

    def process_db_sql(echo=echo):
      typ = 'db_sql'

      for item in self.data[typ]:
        if echo: log('processing item: ' + str(item))
        item_ = prep_item(item)

        if 'sql_list' in item_:
          for n, sql in enumerate(item_['sql_list']):
            name = item['workflow'] + '.' + str(n + 1)
            if name in tasks:
              raise Exception('workflow "{}" already in tasks data!'.format(name))
            else:
              tasks[name] = dict(
                n=n + 1,
                workflow=item['workflow'],
                name=name,
                type=typ,
                database=item['database'],
                sql=sql,
                worker_filter=item['worker_filter'],
                _item=item,
              )
        elif 'sql_file' in item_:
          item_['sql_file'] = item_['sql_file'] if isinstance(item_['sql_file'], list) else [item_['sql_file']]
          for n, file_path in enumerate(item_['sql_file']):
            if not Path(file_path).exists():
              raise Exception('Workflow {} SQL Path "{}" does not exists!'.format(
                item['workflow'], file_path
              ))

            name = '{}.{}.{}'.format(
              item['workflow'], str(n + 1), Path(file_path).name
            )
            tasks[name] = dict(
              n=n + 1,
              workflow=item['workflow'],
              name=name,
              type=typ,
              database=item['database'],
              file_path=file_path,
              worker_filter=item['worker_filter'],
              _item=item,
            )
        else:
          name = item['workflow']
          tasks[name] = dict(
            n=1,
            workflow=item['workflow'],
            name=name,
            type=typ,
            database=item['database'],
            sql=item['sql'],
            worker_filter=item['worker_filter'],
            _item=item,
          )

    def process_bash(echo=echo):
      typ = 'bash'

      for item in self.data[typ]:
        if echo: log('processing item: ' + str(item))
        name = item['workflow']
        tasks[name] = dict(
          n=1,
          name=name,
          workflow=item['workflow'],
          cmd=item['cmd'],
          type=typ,
          _item=item,
        )

    def process_jobs(echo=echo):
      typ = 'jobs'
      jobs = {}
      job_names = list(self.data[typ])
      for job_name in self.data[typ]:
        task_list = self.data[typ][job_name]
        for task_name in task_list:
          if not (task_name in self.task_names or task_name in self.workflow_names or task_name in job_names):
            raise Exception('Task name {} cannot be found!'.format(task_name))
          if job_name in tasks:
            raise Exception('Job name "{}" already in tasks data!'.format(job_name))

        jobs[job_name] = task_list

      return jobs

    if 'db_to_db' in self.data: process_db_to_db(echo=echo)
    if 'ff_to_db' in self.data: process_ff_to_db(echo=echo)
    if 'db_to_ff' in self.data: process_db_to_ff(echo=echo)
    if 'db_sql' in self.data: process_db_sql(echo=echo)
    if 'bash' in self.data: process_bash(echo=echo)

    self.tasks = tasks
    self.task_names = sorted([t for t in tasks])
    self.workflow_names = sorted(set([tasks[t]['workflow'] for t in tasks]))
    self.jobs = process_jobs(echo=echo) if 'jobs' in self.data else {}
    self.job_names = sorted(self.jobs)

  def get_tasks(self):
    sort_key = lambda x: (self.tasks[x]['type'], self.tasks[x]['workflow'], self.tasks[x]['n'])
    tasks = [self.tasks[wn] for wn in sorted(self.tasks, key=sort_key)]
    return tasks


etl_functions = dict(
  db_to_db=db_to_db,
  db_to_ff=db_to_ff,
  ff_to_db=ff_to_db,
  ff_to_ff=ff_to_ff,
  etl_db_sql=exec_sql,
  # bash=bash,
)
